from math import log

import pytest

import wn
from wn import similarity as sim
from wn.taxonomy import taxonomy_depth
from wn.ic import information_content as infocont


def get_synsets(w):
    return {
        "information": w.synset("test-en-0001-n"),
        "example": w.synset("test-en-0002-n"),
        "sample": w.synset("test-en-0004-n"),
        "random sample": w.synset("test-en-0005-n"),
        "random sample2": w.synset("test-en-0008-n"),
        "datum": w.synset("test-en-0006-n"),
        "exemplify": w.synset("test-en-0003-v"),
    }


# some fake information content; computed using:
#   words = ['example', 'example', 'sample', 'random sample', 'illustrate']
#   ic = compute(words, wn.Wordnet('test-en'), distribute_weight=False)

ic = {
    "n": {
        "test-en-0001-n": 5.0,  # information
        "test-en-0002-n": 5.0,  # example, illustration
        "test-en-0004-n": 3.0,  # sample
        "test-en-0005-n": 2.0,  # random sample
        "test-en-0008-n": 2.0,  # random sample 2
        "test-en-0006-n": 1.0,  # datum
        None: 6.0,
    },
    "v": {
        "test-en-0003-v": 2.0,  # exemplify, illustrate
        "test-en-0007-v": 1.0,  # resignate
        None: 2.0,
    },
    "a": {None: 1.0},
    "r": {None: 1.0},
}


@pytest.mark.usefixtures("mini_db")
def test_path():
    ss = get_synsets(wn.Wordnet("test-en"))
    assert sim.path(ss["information"], ss["information"]) == 1 / 1
    assert sim.path(ss["information"], ss["example"]) == 1 / 2
    assert sim.path(ss["information"], ss["sample"]) == 1 / 3
    assert sim.path(ss["information"], ss["random sample"]) == 1 / 4
    assert sim.path(ss["random sample"], ss["datum"]) == 1 / 5
    assert sim.path(ss["random sample2"], ss["datum"]) == 0
    assert sim.path(ss["random sample2"], ss["datum"], simulate_root=True) == 1 / 4
    assert (
        sim.path(ss["random sample"], ss["random sample2"], simulate_root=True) == 1 / 6
    )
    with pytest.raises(wn.Error):
        sim.path(ss["example"], ss["exemplify"])
    with pytest.raises(wn.Error):
        sim.wup(ss["example"], ss["exemplify"], simulate_root=True)


@pytest.mark.usefixtures("mini_db")
def test_wup():
    ss = get_synsets(wn.Wordnet("test-en"))
    assert sim.wup(ss["information"], ss["information"]) == (2 * 1) / (0 + 0 + 2 * 1)
    assert sim.wup(ss["information"], ss["example"]) == (2 * 1) / (0 + 1 + 2 * 1)
    assert sim.wup(ss["information"], ss["sample"]) == (2 * 1) / (0 + 2 + 2 * 1)
    assert sim.wup(ss["information"], ss["random sample"]) == (2 * 1) / (0 + 3 + 2 * 1)
    assert sim.wup(ss["random sample"], ss["datum"]) == (2 * 1) / (3 + 1 + 2 * 1)
    with pytest.raises(wn.Error):
        assert sim.wup(ss["random sample2"], ss["datum"])
    assert sim.wup(ss["random sample2"], ss["datum"], simulate_root=True) == (2 * 1) / (
        1 + 2 + 2 * 1
    )
    assert sim.wup(ss["random sample"], ss["random sample2"], simulate_root=True) == (
        2 * 1
    ) / (4 + 1 + 2 * 1)
    with pytest.raises(wn.Error):
        sim.wup(ss["example"], ss["exemplify"])
    with pytest.raises(wn.Error):
        sim.wup(ss["example"], ss["exemplify"], simulate_root=True)


@pytest.mark.usefixtures("mini_db")
def test_lch():
    w = wn.Wordnet("test-en")
    ss = get_synsets(w)
    d_n = taxonomy_depth(w, "n")
    assert sim.lch(ss["information"], ss["information"], d_n) == -log(
        (0 + 1) / (2 * d_n)
    )
    assert sim.lch(ss["information"], ss["example"], d_n) == -log((1 + 1) / (2 * d_n))
    assert sim.lch(ss["information"], ss["sample"], d_n) == -log((2 + 1) / (2 * d_n))
    assert sim.lch(ss["information"], ss["random sample"], d_n) == -log(
        (3 + 1) / (2 * d_n)
    )
    assert sim.lch(ss["random sample"], ss["datum"], d_n) == -log((4 + 1) / (2 * d_n))
    with pytest.raises(wn.Error):
        assert sim.lch(ss["random sample2"], ss["datum"], d_n)
    assert sim.lch(ss["random sample2"], ss["datum"], d_n, simulate_root=True) == -log(
        (3 + 1) / (2 * d_n)
    )
    assert sim.lch(
        ss["random sample"], ss["random sample2"], d_n, simulate_root=True
    ) == -log((5 + 1) / (2 * d_n))
    with pytest.raises(wn.Error):
        sim.lch(ss["example"], ss["exemplify"], d_n)
    with pytest.raises(wn.Error):
        sim.lch(ss["example"], ss["exemplify"], d_n, simulate_root=True)


@pytest.mark.usefixtures("mini_db")
def test_res():
    w = wn.Wordnet("test-en")
    ss = get_synsets(w)
    assert sim.res(ss["information"], ss["information"], ic) == infocont(
        ss["information"], ic
    )
    assert sim.res(ss["information"], ss["example"], ic) == infocont(
        ss["information"], ic
    )
    assert sim.res(ss["information"], ss["sample"], ic) == infocont(
        ss["information"], ic
    )
    assert sim.res(ss["information"], ss["random sample"], ic) == infocont(
        ss["information"], ic
    )
    assert sim.res(ss["random sample"], ss["datum"], ic) == infocont(
        ss["information"], ic
    )
    with pytest.raises(wn.Error):
        sim.res(ss["random sample2"], ss["datum"], ic)
    with pytest.raises(wn.Error):
        sim.res(ss["example"], ss["exemplify"], ic)


@pytest.mark.usefixtures("mini_db")
def test_jcn():
    w = wn.Wordnet("test-en")
    ss = get_synsets(w)
    info_ic = infocont(ss["information"], ic)
    assert sim.jcn(ss["information"], ss["information"], ic) == float("inf")
    assert sim.jcn(ss["information"], ss["example"], ic) == float("inf")
    assert sim.jcn(ss["information"], ss["sample"], ic) == 1 / (
        (info_ic + infocont(ss["sample"], ic)) - 2 * info_ic
    )
    assert sim.jcn(ss["information"], ss["random sample"], ic) == 1 / (
        (info_ic + infocont(ss["random sample"], ic)) - 2 * info_ic
    )
    assert sim.jcn(ss["random sample"], ss["datum"], ic) == 1 / (
        (infocont(ss["random sample"], ic) + infocont(ss["datum"], ic)) - 2 * info_ic
    )
    with pytest.raises(wn.Error):
        sim.jcn(ss["random sample2"], ss["datum"], ic)
    with pytest.raises(wn.Error):
        sim.jcn(ss["example"], ss["exemplify"], ic)


@pytest.mark.usefixtures("mini_db")
def test_lin():
    w = wn.Wordnet("test-en")
    ss = get_synsets(w)
    info_ic = infocont(ss["information"], ic)
    assert sim.lin(ss["information"], ss["information"], ic) == 1.0
    assert sim.lin(ss["information"], ss["example"], ic) == 1.0
    assert sim.lin(ss["information"], ss["sample"], ic) == (2 * info_ic) / (
        info_ic + infocont(ss["sample"], ic)
    )
    assert sim.lin(ss["information"], ss["random sample"], ic) == (2 * info_ic) / (
        info_ic + infocont(ss["random sample"], ic)
    )
    assert sim.lin(ss["random sample"], ss["datum"], ic) == (
        (2 * info_ic) / (infocont(ss["random sample"], ic) + infocont(ss["datum"], ic))
    )
    with pytest.raises(wn.Error):
        sim.lin(ss["random sample2"], ss["datum"], ic)
    with pytest.raises(wn.Error):
        sim.lin(ss["example"], ss["exemplify"], ic)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "wn"
copyright = "2020, Michael Wayne Goodman"
author = "Michael Wayne Goodman"

import wn

# The short X.Y version
version = ".".join(wn.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags
release = wn.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    # 'sphinx.ext.viewcode',
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Global definitions
rst_prolog = """
.. role:: python(code)
   :language: python
   :class: highlight
"""

# smartquotes = False
smartquotes_action = "De"  # D = en- and em-dash; e = ellipsis

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.#

html_theme = "furo"
html_theme_options = {
    "light_css_variables": {
        "color-brand-primary": "#006699",
        "color-brand-content": "#006699",
        # "color-background": "#f0f0f0",
        # "color-sidebar-background": "#ddd",
    },
    "dark_css_variables": {
        "color-brand-primary": "#00CCFF",
        "color-brand-content": "#00CCFF",
    },
}

html_logo = "_static/wn-logo.svg"

pygments_style = "manni"
pygments_dark_style = "monokai"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = [
    "css/svg.css",
]

# Don't offer to show the source of the current page
html_show_sourcelink = False

# -- Options for autodoc extension -------------------------------------------

autodoc_typehints = "description"
# autodoc_typehints = 'signature'
# autodoc_typehints = 'none'

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/master/", None),
}

# -- Options for sphinx_copybutton extension ---------------------------------

copybutton_prompt_text = (
    r">>> "  # regular Python prompt
    r"|\.\.\. "  # Python continuation prompt
    r"|\$ "  # Basic shell
    r"|In \[\d*\]: "  # Jupyter notebook
)
copybutton_prompt_is_regexp = True



import pytest

import wn


@pytest.mark.usefixtures("empty_db")
def test_lexicons_empty():
    assert len(wn.lexicons()) == 0


@pytest.mark.usefixtures("mini_db")
def test_lexicons_mini():
    assert len(wn.lexicons()) == 2
    assert all(isinstance(lex, wn.Lexicon) for lex in wn.lexicons())

    results = wn.lexicons(lang="en")
    assert len(results) == 1 and results[0].language == "en"
    results = wn.lexicons(lang="es")
    assert len(results) == 1 and results[0].language == "es"

    assert len(wn.lexicons(lexicon="*")) == 2
    assert len(wn.lexicons(lexicon="*:1")) == 2
    assert len(wn.lexicons(lexicon="test-*")) == 2
    assert len(wn.lexicons(lexicon="*-en")) == 1
    results = wn.lexicons(lexicon="test-en")
    assert len(results) == 1 and results[0].language == "en"
    results = wn.lexicons(lexicon="test-en:1")
    assert len(results) == 1 and results[0].language == "en"
    results = wn.lexicons(lexicon="test-en:*")
    assert len(results) == 1 and results[0].language == "en"

    assert wn.lexicons(lexicon="test-en")[0].specifier() == "test-en:1"
    assert wn.lexicons(lexicon="test-es")[0].specifier() == "test-es:1"

    assert wn.lexicons(lexicon="test-en")[0].requires() == {}
    assert wn.lexicons(lexicon="test-es")[0].requires() == {}


@pytest.mark.usefixtures("mini_db")
def test_lexicons_unknown():
    results = wn.lexicons(lang="unk")
    assert len(results) == 0
    results = wn.lexicons(lexicon="test-unk")
    assert len(results) == 0


@pytest.mark.usefixtures("empty_db")
def test_words_empty():
    assert len(wn.words()) == 0


@pytest.mark.usefixtures("mini_db")
def test_words_mini():
    assert len(wn.words()) == 15
    assert all(isinstance(w, wn.Word) for w in wn.words())

    words = wn.words("information")  # search lemma
    assert len(words) == 1
    assert words[0].lemma() == "information"

    assert words[0].lemma().script == "Latn"
    assert words[0].lemma().tags() == [wn.Tag("tag-text", "tag-category")]

    words = wn.words("exemplifies")  # search secondary form
    assert len(words) == 1
    assert words[0].lemma() == "exemplify"

    assert len(wn.words(pos="n")) == 10
    assert all(w.pos == "n" for w in wn.words(pos="n"))
    assert len(wn.words(pos="v")) == 5
    assert len(wn.words(pos="q")) == 0  # fake pos

    assert len(wn.words(lang="en")) == 9
    assert len(wn.words(lang="es")) == 6

    assert len(wn.words(lexicon="test-en")) == 9
    assert len(wn.words(lexicon="test-es")) == 6

    assert len(wn.words(lang="en", lexicon="test-en")) == 9
    assert len(wn.words(pos="v", lang="en")) == 3
    assert len(wn.words("information", lang="en")) == 1
    assert len(wn.words("information", lang="es")) == 0

    with pytest.raises(wn.Error):
        wn.words(lang="unk")
    with pytest.raises(wn.Error):
        wn.words(lexicon="test-unk")


@pytest.mark.usefixtures("empty_db")
def test_word_empty():
    with pytest.raises(wn.Error):
        assert wn.word("test-es-información-n")


@pytest.mark.usefixtures("mini_db")
def test_word_mini():
    assert wn.word("test-es-información-n")
    assert wn.word("test-es-información-n", lang="es")
    assert wn.word("test-es-información-n", lexicon="test-es")
    with pytest.raises(wn.Error):
        assert wn.word("test-es-información-n", lang="en")
    with pytest.raises(wn.Error):
        assert wn.word("test-es-información-n", lexicon="test-en")
    with pytest.raises(wn.Error):
        assert wn.word("test-es-información-n", lang="unk")
    with pytest.raises(wn.Error):
        assert wn.word("test-es-información-n", lexicon="test-unk")


@pytest.mark.usefixtures("empty_db")
def test_senses_empty():
    assert len(wn.senses()) == 0


@pytest.mark.usefixtures("mini_db")
def test_senses_mini():
    assert len(wn.senses()) == 16
    assert all(isinstance(s, wn.Sense) for s in wn.senses())

    senses = wn.senses("information")  # search lemma
    assert len(senses) == 1
    assert senses[0].word().lemma() == "information"
    assert senses[0].counts() == [3]

    senses = wn.senses("exemplifies")  # search secondary form
    assert len(senses) == 1
    assert senses[0].word().lemma() == "exemplify"
    assert senses[0].word().lemma() in {"exemplify"}
    assert "exemplify" in {senses[0].word().lemma()}

    assert len(wn.senses(pos="n")) == 11
    assert len(wn.senses(pos="v")) == 5
    assert len(wn.senses(pos="q")) == 0  # fake pos

    assert len(wn.senses(lang="en")) == 10
    assert len(wn.senses(lang="es")) == 6

    assert len(wn.senses(lexicon="test-en")) == 10
    assert len(wn.senses(lexicon="test-es")) == 6

    assert len(wn.senses(lang="en", lexicon="test-en")) == 10
    assert len(wn.senses(pos="v", lang="en")) == 3
    assert len(wn.senses("information", lang="en")) == 1
    assert len(wn.senses("information", lang="es")) == 0

    with pytest.raises(wn.Error):
        wn.senses(lang="unk")
    with pytest.raises(wn.Error):
        wn.senses(lexicon="test-unk")


@pytest.mark.usefixtures("empty_db")
def test_sense_empty():
    with pytest.raises(wn.Error):
        assert wn.sense("test-es-información-n-0001-01")


@pytest.mark.usefixtures("mini_db")
def test_sense_mini():
    assert wn.sense("test-es-información-n-0001-01")
    assert wn.sense("test-es-información-n-0001-01", lang="es")
    assert wn.sense("test-es-información-n-0001-01", lexicon="test-es")
    with pytest.raises(wn.Error):
        assert wn.sense("test-es-información-n-0001-01", lang="en")
    with pytest.raises(wn.Error):
        assert wn.sense("test-es-información-n-0001-01", lexicon="test-en")
    with pytest.raises(wn.Error):
        assert wn.sense("test-es-información-n-0001-01", lang="unk")
    with pytest.raises(wn.Error):
        assert wn.sense("test-es-información-n-0001-01", lexicon="test-unk")


@pytest.mark.usefixtures("empty_db")
def test_synsets_empty():
    assert len(wn.synsets()) == 0


@pytest.mark.usefixtures("mini_db")
def test_synsets_mini():
    assert len(wn.synsets()) == 12
    assert all(isinstance(ss, wn.Synset) for ss in wn.synsets())

    synsets = wn.synsets("information")  # search lemma
    assert len(synsets) == 1
    assert "information" in synsets[0].lemmas()

    synsets = wn.synsets("exemplifies")  # search secondary form
    assert len(synsets) == 1
    assert "exemplify" in synsets[0].lemmas()

    assert len(wn.synsets(pos="n")) == 9
    assert len(wn.synsets(pos="v")) == 3
    assert len(wn.synsets(pos="q")) == 0  # fake pos

    assert len(wn.synsets(ili="i67469")) == 2
    assert len(wn.synsets(ili="i67468")) == 0

    assert len(wn.synsets(lang="en")) == 8
    assert len(wn.synsets(lang="es")) == 4

    assert len(wn.synsets(lexicon="test-en")) == 8
    assert len(wn.synsets(lexicon="test-es")) == 4

    assert len(wn.synsets(lang="en", lexicon="test-en")) == 8
    assert len(wn.synsets(pos="v", lang="en")) == 2
    assert len(wn.synsets("information", lang="en")) == 1
    assert len(wn.synsets("information", lang="es")) == 0
    assert len(wn.synsets(ili="i67469", lang="es")) == 1

    with pytest.raises(wn.Error):
        wn.synsets(lang="unk")
    with pytest.raises(wn.Error):
        wn.synsets(lexicon="test-unk")


@pytest.mark.usefixtures("empty_db")
def test_synset_empty():
    with pytest.raises(wn.Error):
        assert wn.synset("test-es-0001-n")


@pytest.mark.usefixtures("mini_db")
def test_synset_mini():
    assert wn.synset("test-es-0001-n")
    assert wn.synset("test-es-0001-n", lang="es")
    assert wn.synset("test-es-0001-n", lexicon="test-es")
    with pytest.raises(wn.Error):
        assert wn.synset("test-es-0001-n", lang="en")
    with pytest.raises(wn.Error):
        assert wn.synset("test-es-0001-n", lexicon="test-en")
    with pytest.raises(wn.Error):
        assert wn.synset("test-es-0001-n", lang="unk")
    with pytest.raises(wn.Error):
        assert wn.synset("test-es-0001-n", lexicon="test-unk")


@pytest.mark.usefixtures("mini_db_1_1")
def test_mini_1_1():
    assert len(wn.lexicons()) == 4
    assert len(wn.lexicons(lang="en")) == 2
    assert len(wn.lexicons(lang="ja")) == 1
    assert wn.lexicons(lang="ja")[0].logo == "logo.svg"

    w = wn.Wordnet(lang="en")
    assert len(w.lexicons()) == 2
    assert len(w.expanded_lexicons()) == 0
    assert len(w.word("test-en-exemplify-v").lemma().tags()) == 1

    w = wn.Wordnet(lang="ja")
    assert len(w.lexicons()) == 1
    assert len(w.expanded_lexicons()) == 1
    assert len(w.synsets("例え")[0].hypernyms()) == 1
    assert w.synsets("例え")[0].lexfile() == "noun.cognition"
    assert len(w.word("test-ja-例え-n").lemma().pronunciations()) == 1
    assert w.word("test-ja-例え-n").forms()[1].id == "test-ja-例え-n-たとえ"
    p = w.word("test-ja-例え-n").lemma().pronunciations()[0]
    assert p.value == "tatoe"
    assert p.variety == "standard"
    assert p.notation == "ipa"
    assert p.phonemic
    assert p.audio == "tatoe.wav"

    w = wn.Wordnet(lang="ja", expand="")
    assert len(w.lexicons()) == 1
    assert len(w.expanded_lexicons()) == 0
    assert len(w.synsets("例え")[0].hypernyms()) == 0

    w = wn.Wordnet(lexicon="test-en test-en-ext")
    assert len(w.lexicons()) == 2
    assert len(w.expanded_lexicons()) == 0
    assert len(w.synsets("fire")[0].hyponyms()) == 1



import pytest

import wn


@pytest.mark.usefixtures("mini_db")
def test_word_senses():
    assert len(wn.word("test-en-information-n").senses()) == 1
    assert len(wn.word("test-es-información-n").senses()) == 1


@pytest.mark.usefixtures("mini_db")
def test_word_synsets():
    assert len(wn.word("test-en-information-n").synsets()) == 1
    assert len(wn.word("test-es-información-n").synsets()) == 1


@pytest.mark.usefixtures("mini_db")
def test_word_translate():
    assert len(wn.word("test-en-example-n").translate(lang="es")) == 1
    assert len(wn.word("test-es-ejemplo-n").translate(lang="en")) == 1


@pytest.mark.usefixtures("mini_db")
def test_sense_word():
    assert wn.sense("test-en-information-n-0001-01").word() == wn.word(
        "test-en-information-n"
    )
    assert wn.sense("test-es-información-n-0001-01").word() == wn.word(
        "test-es-información-n"
    )


@pytest.mark.usefixtures("mini_db")
def test_sense_synset():
    assert wn.sense("test-en-information-n-0001-01").synset() == wn.synset(
        "test-en-0001-n"
    )
    assert wn.sense("test-es-información-n-0001-01").synset() == wn.synset(
        "test-es-0001-n"
    )


@pytest.mark.usefixtures("mini_db")
def test_sense_examples():
    assert wn.sense("test-en-information-n-0001-01").examples() == []
    assert wn.sense("test-es-información-n-0001-01").examples() == []


@pytest.mark.usefixtures("mini_db")
def test_sense_lexicalized():
    assert wn.sense("test-en-information-n-0001-01").lexicalized()
    assert wn.sense("test-es-información-n-0001-01").lexicalized()


@pytest.mark.usefixtures("mini_db")
def test_sense_frames():
    assert wn.sense("test-en-illustrate-v-0003-01").frames() == [
        "Somebody ----s something",
        "Something ----s something",
    ]
    assert wn.sense("test-es-ilustrar-v-0003-01").frames() == []


@pytest.mark.usefixtures("mini_db_1_1")
def test_sense_frames_issue_156():
    # https://github.com/goodmami/wn/issues/156
    assert wn.sense("test-ja-示す-v-0003-01").frames() == [
        "ある人が何かを----",
    ]
    assert wn.sense("test-ja-事例-n-0002-01").frames() == []


@pytest.mark.usefixtures("mini_db")
def test_sense_translate():
    assert len(wn.sense("test-en-information-n-0001-01").translate(lang="es")) == 1
    assert len(wn.sense("test-es-información-n-0001-01").translate(lang="en")) == 1


@pytest.mark.usefixtures("mini_db")
def test_synset_senses():
    assert len(wn.synset("test-en-0003-v").senses()) == 2
    assert len(wn.synset("test-es-0003-v").senses()) == 2


@pytest.mark.usefixtures("mini_db")
def test_synset_words():
    assert len(wn.synset("test-en-0003-v").words()) == 2
    assert len(wn.synset("test-es-0003-v").words()) == 2


@pytest.mark.usefixtures("mini_db")
def test_synset_lemmas():
    assert wn.synset("test-en-0003-v").lemmas() == ["exemplify", "illustrate"]
    assert wn.synset("test-es-0003-v").lemmas() == ["ejemplificar", "ilustrar"]


@pytest.mark.usefixtures("mini_db")
def test_synset_ili():
    assert isinstance(wn.synset("test-en-0001-n").ili, wn.ILI)
    assert wn.synset("test-en-0001-n").ili.id == "i67447"
    assert wn.synset("test-en-0001-n").ili.status == "presupposed"
    assert wn.synset("test-en-0008-n").ili is None
    assert wn.synset("test-en-0007-v").ili.id is None
    assert wn.synset("test-en-0007-v").ili.status == "proposed"


@pytest.mark.usefixtures("mini_db")
def test_synset_definition():
    assert wn.synset("test-en-0001-n").definition() == "something that informs"
    assert wn.synset("test-es-0001-n").definition() == "algo que informa"


@pytest.mark.usefixtures("mini_db")
def test_synset_examples():
    assert wn.synset("test-en-0001-n").examples() == ['"this is information"']
    assert wn.synset("test-es-0001-n").examples() == ['"este es la información"']


@pytest.mark.usefixtures("mini_db")
def test_synset_lexicalized():
    assert wn.synset("test-en-0001-n").lexicalized()
    assert wn.synset("test-es-0001-n").lexicalized()


@pytest.mark.usefixtures("mini_db")
def test_synset_translate():
    assert len(wn.synset("test-en-0001-n").translate(lang="es")) == 1
    assert len(wn.synset("test-es-0001-n").translate(lang="en")) == 1



from pathlib import Path
import tempfile

import pytest

import wn


@pytest.fixture(scope="session")
def datadir():
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def mini_lmf_1_0(datadir):
    return datadir / "mini-lmf-1.0.xml"


@pytest.fixture(scope="session")
def mini_lmf_1_1(datadir):
    return datadir / "mini-lmf-1.1.xml"


@pytest.fixture(scope="session")
def empty_db_dir():
    with tempfile.TemporaryDirectory("wn_data_empty") as dir:
        yield Path(dir)


@pytest.fixture(scope="session")
def mini_db_dir(mini_lmf_1_0):
    with tempfile.TemporaryDirectory("wn_data_mini") as dir:
        old_data_dir = wn.config.data_directory
        wn.config.data_directory = dir
        wn.add(mini_lmf_1_0)
        wn.config.data_directory = old_data_dir
        yield Path(dir)
        # close any open DB connections before teardown
        for conn in wn._db.pool.values():
            conn.close()


@pytest.fixture(scope="session")
def mini_db_1_1_dir(mini_lmf_1_0, mini_lmf_1_1):
    with tempfile.TemporaryDirectory("wn_data_mini_1_1") as dir:
        old_data_dir = wn.config.data_directory
        wn.config.data_directory = dir
        wn.add(mini_lmf_1_0)
        wn.add(mini_lmf_1_1)
        wn.config.data_directory = old_data_dir
        yield Path(dir)
        # close any open DB connections before teardown
        for conn in wn._db.pool.values():
            conn.close()


@pytest.fixture
def empty_db(monkeypatch, empty_db_dir):
    with monkeypatch.context() as m:
        m.setattr(wn.config, "data_directory", empty_db_dir)
        yield


@pytest.fixture
def mini_db(monkeypatch, mini_db_dir):
    with monkeypatch.context() as m:
        m.setattr(wn.config, "data_directory", mini_db_dir)
        yield


@pytest.fixture
def mini_db_1_1(monkeypatch, mini_db_1_1_dir):
    with monkeypatch.context() as m:
        m.setattr(wn.config, "data_directory", mini_db_1_1_dir)
        yield



from xml.etree import ElementTree as ET

from wn import lmf


def test_is_lmf(datadir):
    assert lmf.is_lmf(datadir / "mini-lmf-1.0.xml")
    assert lmf.is_lmf(str(datadir / "mini-lmf-1.0.xml"))
    assert not lmf.is_lmf(datadir / "README.md")
    assert not lmf.is_lmf(datadir / "missing.xml")
    assert lmf.is_lmf(datadir / "mini-lmf-1.1.xml")


def test_load_1_0(mini_lmf_1_0):
    resource = lmf.load(mini_lmf_1_0)
    lexicons = resource["lexicons"]
    assert len(lexicons) == 2
    lexicon = lexicons[0]

    assert lexicon["id"] == "test-en"
    assert lexicon["label"] == "Testing English WordNet"
    assert lexicon["language"] == "en"
    assert lexicon["email"] == "maintainer@example.com"
    assert lexicon["license"] == "https://creativecommons.org/licenses/by/4.0/"
    assert lexicon["version"] == "1"
    assert lexicon["url"] == "https://example.com/test-en"

    assert len(lexicon["entries"]) == 9
    le = lexicon["entries"][0]
    assert le["id"] == "test-en-information-n"

    assert le["lemma"]["writtenForm"] == "information"
    assert le["lemma"]["partOfSpeech"] == "n"
    assert le["lemma"]["script"] == "Latn"
    assert len(le["lemma"]["tags"]) == 1

    assert len(le.get("forms", [])) == 0

    assert len(le["senses"]) == 1
    sense = le["senses"][0]
    assert sense["id"] == "test-en-information-n-0001-01"
    assert sense["synset"] == "test-en-0001-n"
    assert len(sense.get("relations", [])) == 0
    # assert sense['relations'][0]['target'] == 'test-en-exemplify-v-01023137-01'
    # assert sense['relations'][0]['type'] == 'derivation'

    assert len(lexicon.get("frames", [])) == 0  # frames are on lexical entry
    assert len(lexicon["entries"][6]["frames"]) == 2
    frames = lexicon["entries"][6]["frames"]
    assert frames[0]["subcategorizationFrame"] == "Somebody ----s something"
    assert frames[0]["senses"] == ["test-en-illustrate-v-0003-01"]

    assert len(lexicon["synsets"]) == 8

    assert lexicons[1]["id"] == "test-es"


def test_load_1_1(mini_lmf_1_1):
    resource = lmf.load(mini_lmf_1_1)
    lexicons = resource["lexicons"]
    assert len(lexicons) == 2
    lexicon = lexicons[0]
    assert lexicon["id"] == "test-ja"
    assert lexicon["version"] == "1"
    # assert lexicon.logo == 'logo.svg'
    assert lexicon.get("requires") == [{"id": "test-en", "version": "1"}]

    lexicon = lexicons[1]
    assert lexicon["id"] == "test-en-ext"
    assert lexicon.get("extends") == {"id": "test-en", "version": "1"}


def test_dump(mini_lmf_1_0, mini_lmf_1_1, tmp_path):
    tmpdir = tmp_path / "test_dump"
    tmpdir.mkdir()
    tmppath = tmpdir / "mini_lmf_dump.xml"

    def assert_xml_equal(mini_lmf, dump_lmf):
        if hasattr(ET, "canonicalize"):  # available from Python 3.8
            orig = ET.canonicalize(from_file=mini_lmf, strip_text=True)
            temp = ET.canonicalize(from_file=dump_lmf, strip_text=True)
            # additional transformation to help with debugging
            orig = orig.replace("<", "\n<")
            temp = temp.replace("<", "\n<")
            assert orig == temp

    lmf.dump(lmf.load(mini_lmf_1_0), tmppath)
    assert_xml_equal(mini_lmf_1_0, tmppath)

    lmf.dump(lmf.load(mini_lmf_1_1), tmppath)
    assert_xml_equal(mini_lmf_1_1, tmppath)



"""
Constants and literals used in wordnets.
"""


SENSE_RELATIONS = frozenset(
    [
        "antonym",
        "also",
        "participle",
        "pertainym",
        "derivation",
        "domain_topic",
        "has_domain_topic",
        "domain_region",
        "has_domain_region",
        "exemplifies",
        "is_exemplified_by",
        "similar",
        "other",
        "feminine",
        "has_feminine",
        "masculine",
        "has_masculine",
        "young",
        "has_young",
        "diminutive",
        "has_diminutive",
        "augmentative",
        "has_augmentative",
        "anto_gradable",
        "anto_simple",
        "anto_converse",
        "simple_aspect_ip",
        "secondary_aspect_ip",
        "simple_aspect_pi",
        "secondary_aspect_pi",
    ]
)

SENSE_SYNSET_RELATIONS = frozenset(
    [
        "other",
        "domain_topic",
        "domain_region",
        "exemplifies",
    ]
)

SYNSET_RELATIONS = frozenset(
    [
        "agent",
        "also",
        "attribute",
        "be_in_state",
        "causes",
        "classified_by",
        "classifies",
        "co_agent_instrument",
        "co_agent_patient",
        "co_agent_result",
        "co_instrument_agent",
        "co_instrument_patient",
        "co_instrument_result",
        "co_patient_agent",
        "co_patient_instrument",
        "co_result_agent",
        "co_result_instrument",
        "co_role",
        "direction",
        "domain_region",
        "domain_topic",
        "exemplifies",
        "entails",
        "eq_synonym",
        "has_domain_region",
        "has_domain_topic",
        "is_exemplified_by",
        "holo_location",
        "holo_member",
        "holo_part",
        "holo_portion",
        "holo_substance",
        "holonym",
        "hypernym",
        "hyponym",
        "in_manner",
        "instance_hypernym",
        "instance_hyponym",
        "instrument",
        "involved",
        "involved_agent",
        "involved_direction",
        "involved_instrument",
        "involved_location",
        "involved_patient",
        "involved_result",
        "involved_source_direction",
        "involved_target_direction",
        "is_caused_by",
        "is_entailed_by",
        "location",
        "manner_of",
        "mero_location",
        "mero_member",
        "mero_part",
        "mero_portion",
        "mero_substance",
        "meronym",
        "similar",
        "other",
        "patient",
        "restricted_by",
        "restricts",
        "result",
        "role",
        "source_direction",
        "state_of",
        "target_direction",
        "subevent",
        "is_subevent_of",
        "antonym",
        "feminine",
        "has_feminine",
        "masculine",
        "has_masculine",
        "young",
        "has_young",
        "diminutive",
        "has_diminutive",
        "augmentative",
        "has_augmentative",
        "anto_gradable",
        "anto_simple",
        "anto_converse",
        "ir_synonym",
    ]
)


REVERSE_RELATIONS = {
    "hypernym": "hyponym",
    "hyponym": "hypernym",
    "instance_hypernym": "instance_hyponym",
    "instance_hyponym": "instance_hypernym",
    "antonym": "antonym",
    "eq_synonym": "eq_synonym",
    "similar": "similar",
    "meronym": "holonym",
    "holonym": "meronym",
    "mero_location": "holo_location",
    "holo_location": "mero_location",
    "mero_member": "holo_member",
    "holo_member": "mero_member",
    "mero_part": "holo_part",
    "holo_part": "mero_part",
    "mero_portion": "holo_portion",
    "holo_portion": "mero_portion",
    "mero_substance": "holo_substance",
    "holo_substance": "mero_substance",
    "also": "also",
    "state_of": "be_in_state",
    "be_in_state": "state_of",
    "causes": "is_caused_by",
    "is_caused_by": "causes",
    "subevent": "is_subevent_of",
    "is_subevent_of": "subevent",
    "manner_of": "in_manner",
    "in_manner": "manner_of",
    "attribute": "attribute",
    "restricts": "restricted_by",
    "restricted_by": "restricts",
    "classifies": "classified_by",
    "classified_by": "classifies",
    "entails": "is_entailed_by",
    "is_entailed_by": "entails",
    "domain_topic": "has_domain_topic",
    "has_domain_topic": "domain_topic",
    "domain_region": "has_domain_region",
    "has_domain_region": "domain_region",
    "exemplifies": "is_exemplified_by",
    "is_exemplified_by": "exemplifies",
    "role": "involved",
    "involved": "role",
    "agent": "involved_agent",
    "involved_agent": "agent",
    "patient": "involved_patient",
    "involved_patient": "patient",
    "result": "involved_result",
    "involved_result": "result",
    "instrument": "involved_instrument",
    "involved_instrument": "instrument",
    "location": "involved_location",
    "involved_location": "location",
    "direction": "involved_direction",
    "involved_direction": "direction",
    "target_direction": "involved_target_direction",
    "involved_target_direction": "target_direction",
    "source_direction": "involved_source_direction",
    "involved_source_direction": "source_direction",
    "co_role": "co_role",
    "co_agent_patient": "co_patient_agent",
    "co_patient_agent": "co_agent_patient",
    "co_agent_instrument": "co_instrument_agent",
    "co_instrument_agent": "co_agent_instrument",
    "co_agent_result": "co_result_agent",
    "co_result_agent": "co_agent_result",
    "co_patient_instrument": "co_instrument_patient",
    "co_instrument_patient": "co_patient_instrument",
    "co_result_instrument": "co_instrument_result",
    "co_instrument_result": "co_result_instrument",
    "pertainym": "pertainym",
    "derivation": "derivation",
    "simple_aspect_ip": "simple_aspect_pi",
    "simple_aspect_pi": "simple_aspect_ip",
    "secondary_aspect_ip": "secondary_aspect_pi",
    "secondary_aspect_pi": "secondary_aspect_ip",
    "feminine": "has_feminine",
    "has_feminine": "feminine",
    "masculine": "has_masculine",
    "has_masculine": "masculine",
    "young": "has_young",
    "has_young": "young",
    "diminutive": "has_diminutive",
    "has_diminutive": "diminutive",
    "augmentative": "has_augmentative",
    "has_augmentative": "augmentative",
    "anto_gradable": "anto_gradable",
    "anto_simple": "anto_simple",
    "anto_converse": "anto_converse",
    "ir_synonym": "ir_synonym",
    # 'participle': '',
    # 'other': '',
}

# Adjective Positions

ADJPOSITIONS = frozenset(
    (
        "a",  # attributive
        "ip",  # immediate postnominal
        "p",  # predicative
    )
)


# Parts of Speech

NOUN = "n"  #:
VERB = "v"  #:
ADJ = ADJECTIVE = "a"  #:
ADV = ADVERB = "r"  #:
ADJ_SAT = ADJECTIVE_SATELLITE = "s"  #:
PHRASE = "t"  #:
CONJ = CONJUNCTION = "c"  #:
ADP = ADPOSITION = "p"  #:
OTHER = "x"  #:
UNKNOWN = "u"  #:

PARTS_OF_SPEECH = frozenset(
    (
        NOUN,
        VERB,
        ADJECTIVE,
        ADVERB,
        ADJECTIVE_SATELLITE,
        PHRASE,
        CONJUNCTION,
        ADPOSITION,
        OTHER,
        UNKNOWN,
    )
)


# Lexicographer Files
# from https://wordnet.princeton.edu/documentation/lexnames5wn

LEXICOGRAPHER_FILES = {
    "adj.all": 0,
    "adj.pert": 1,
    "adv.all": 2,
    "noun.Tops": 3,
    "noun.act": 4,
    "noun.animal": 5,
    "noun.artifact": 6,
    "noun.attribute": 7,
    "noun.body": 8,
    "noun.cognition": 9,
    "noun.communication": 10,
    "noun.event": 11,
    "noun.feeling": 12,
    "noun.food": 13,
    "noun.group": 14,
    "noun.location": 15,
    "noun.motive": 16,
    "noun.object": 17,
    "noun.person": 18,
    "noun.phenomenon": 19,
    "noun.plant": 20,
    "noun.possession": 21,
    "noun.process": 22,
    "noun.quantity": 23,
    "noun.relation": 24,
    "noun.shape": 25,
    "noun.state": 26,
    "noun.substance": 27,
    "noun.time": 28,
    "verb.body": 29,
    "verb.change": 30,
    "verb.cognition": 31,
    "verb.communication": 32,
    "verb.competition": 33,
    "verb.consumption": 34,
    "verb.contact": 35,
    "verb.creation": 36,
    "verb.emotion": 37,
    "verb.motion": 38,
    "verb.perception": 39,
    "verb.possession": 40,
    "verb.social": 41,
    "verb.stative": 42,
    "verb.weather": 43,
    "adj.ppl": 44,
}

# resource types

_WORDNET = "wordnet"
_ILI = "ili"



"""Information Content is a corpus-based metrics of synset or sense
specificity.

"""

from typing import (
    Callable,
    Optional,
    Iterator,
    Iterable,
    Dict,
    List,
    Tuple,
    Set,
    TextIO,
)
from pathlib import Path
from collections import Counter
from math import log

from wn._types import AnyPath
from wn._core import Synset, Wordnet
from wn.constants import NOUN, VERB, ADJ, ADV, ADJ_SAT
from wn.util import synset_id_formatter


# Just use a subset of all available parts of speech
IC_PARTS_OF_SPEECH = frozenset((NOUN, VERB, ADJ, ADV))
Freq = Dict[str, Dict[Optional[str], float]]


def information_content(synset: Synset, freq: Freq) -> float:
    """Calculate the Information Content value for a synset.

    The information content of a synset is the negative log of the
    synset probability (see :func:`synset_probability`).

    """
    return -log(synset_probability(synset, freq))


def synset_probability(synset: Synset, freq: Freq) -> float:
    """Calculate the synset probability.

    The synset probability is defined as freq(ss)/N where freq(ss) is
    the IC weight for the synset and N is the total IC weight for all
    synsets with the same part of speech.

    Note: this function is not generally used directly, but indirectly
    through :func:`information_content`.

    """
    pos_freq = freq[synset.pos]
    return pos_freq[synset.id] / pos_freq[None]


def _initialize(
    wordnet: Wordnet,
    smoothing: float,
) -> Freq:
    """Populate an Information Content weight mapping to a smoothing value.

    All synsets in *wordnet* are inserted into the dictionary and
    mapped to *smoothing*.

    """
    freq: Freq = {
        pos: {synset.id: smoothing for synset in wordnet.synsets(pos=pos)}
        for pos in IC_PARTS_OF_SPEECH
    }
    # pretend ADJ_SAT is just ADJ
    for synset in wordnet.synsets(pos=ADJ_SAT):
        freq[ADJ][synset.id] = smoothing
    # also initialize totals (when synset is None) for each part-of-speech
    for pos in IC_PARTS_OF_SPEECH:
        freq[pos][None] = smoothing
    return freq


def compute(
    corpus: Iterable[str],
    wordnet: Wordnet,
    distribute_weight: bool = True,
    smoothing: float = 1.0,
) -> Freq:
    """Compute Information Content weights from a corpus.

    Arguments:
        corpus: An iterable of string tokens. This is a flat list of
            words and the order does not matter. Tokens may be single
            words or multiple words separated by a space.

        wordnet: An instantiated :class:`wn.Wordnet` object, used to
            look up synsets from words.

        distribute_weight: If :python:`True`, the counts for a word
            are divided evenly among all synsets for the word.

        smoothing: The initial value given to each synset.

    Example:
        >>> import wn, wn.ic, wn.morphy
        >>> ewn = wn.Wordnet('ewn:2020', lemmatizer=wn.morphy.morphy)
        >>> freq = wn.ic.compute(["Dogs", "run", ".", "Cats", "sleep", "."], ewn)
        >>> dog = ewn.synsets('dog', pos='n')[0]
        >>> cat = ewn.synsets('cat', pos='n')[0]
        >>> frog = ewn.synsets('frog', pos='n')[0]
        >>> freq['n'][dog.id]
        1.125
        >>> freq['n'][cat.id]
        1.1
        >>> freq['n'][frog.id]  # no occurrence; smoothing value only
        1.0
        >>> carnivore = dog.lowest_common_hypernyms(cat)[0]
        >>> freq['n'][carnivore.id]
        1.3250000000000002
    """
    freq = _initialize(wordnet, smoothing)
    counts = Counter(corpus)

    hypernym_cache: Dict[Synset, List[Synset]] = {}
    for word, count in counts.items():
        synsets = wordnet.synsets(word)
        num = len(synsets)
        if num == 0:
            continue

        weight = float(count / num if distribute_weight else count)

        for synset in synsets:
            pos = synset.pos
            if pos == ADJ_SAT:
                pos = ADJ
            if pos not in IC_PARTS_OF_SPEECH:
                continue

            freq[pos][None] += weight

            # The following while-loop is equivalent to:
            #
            # freq[pos][synset.id] += weight
            # for path in synset.hypernym_paths():
            #     for ss in path:
            #         freq[pos][ss.id] += weight
            #
            # ...but it caches hypernym lookups for speed

            agenda: List[Tuple[Synset, Set[Synset]]] = [(synset, set())]
            while agenda:
                ss, seen = agenda.pop()

                # avoid cycles
                if ss in seen:
                    continue

                freq[pos][ss.id] += weight

                if ss not in hypernym_cache:
                    hypernym_cache[ss] = ss.hypernyms()
                agenda.extend((hyp, seen | {ss}) for hyp in hypernym_cache[ss])

    return freq


def load(
    source: AnyPath,
    wordnet: Wordnet,
    get_synset_id: Optional[Callable] = None,
) -> Freq:
    """Load an Information Content mapping from a file.

    Arguments:

        source: A path to an information content weights file.

        wordnet: A :class:`wn.Wordnet` instance with synset
            identifiers matching the offsets in the weights file.

        get_synset_id: A callable that takes a synset offset and part
            of speech and returns a synset ID valid in *wordnet*.

    Raises:

        :class:`wn.Error`: If *wordnet* does not have exactly one
            lexicon.

    Example:

        >>> import wn, wn.ic
        >>> pwn = wn.Wordnet('pwn:3.0')
        >>> path = '~/nltk_data/corpora/wordnet_ic/ic-brown-resnik-add1.dat'
        >>> freq = wn.ic.load(path, pwn)

    """
    source = Path(source).expanduser().resolve(strict=True)
    assert len(wordnet.lexicons()) == 1
    lexid = wordnet.lexicons()[0].id
    if get_synset_id is None:
        get_synset_id = synset_id_formatter(prefix=lexid)

    freq = _initialize(wordnet, 0.0)

    with source.open() as icfile:
        for offset, pos, weight, is_root in _parse_ic_file(icfile):
            ssid = get_synset_id(offset=offset, pos=pos)
            # synset = wordnet.synset(ssid)
            freq[pos][ssid] = weight
            if is_root:
                freq[pos][None] += weight
    return freq


def _parse_ic_file(icfile: TextIO) -> Iterator[Tuple[int, str, float, bool]]:
    """Parse the Information Content file.

    A sample of the format is::

        wnver::eOS9lXC6GvMWznF1wkZofDdtbBU
        1740n 1915712 ROOT
        1930n 859272
        2137n 1055337

    """
    next(icfile)  # skip header
    for line in icfile:
        ssinfo, value, *isroot = line.split()
        yield (int(ssinfo[:-1]), ssinfo[-1], float(value), bool(isroot))



import pytest

import wn
from wn.taxonomy import (
    roots,
    leaves,
    taxonomy_depth,
    hypernym_paths,
    min_depth,
    max_depth,
    shortest_path,
    # common_hypernyms,
    # lowest_common_hypernyms,
)


@pytest.mark.usefixtures("mini_db")
def test_roots():
    en = wn.Wordnet("test-en")
    assert set(roots(en, pos="n")) == {
        en.synset("test-en-0001-n"),
        en.synset("test-en-0008-n"),
    }
    assert set(roots(en, pos="v")) == {
        en.synset("test-en-0003-v"),
        en.synset("test-en-0007-v"),
    }
    assert roots(en, pos="a") == []
    assert set(roots(en)) == set(roots(en, pos="n") + roots(en, pos="v"))

    # with no expand relations and no relation of its own, every
    # synset looks like a root
    es = wn.Wordnet("test-es")
    assert set(roots(es, pos="n")) == {
        es.synset("test-es-0001-n"),
        es.synset("test-es-0002-n"),
        es.synset("test-es-0005-n"),
    }

    es = wn.Wordnet("test-es", expand="test-en")
    assert roots(es, pos="n") == [es.synset("test-es-0001-n")]


@pytest.mark.usefixtures("mini_db")
def test_leaves():
    en = wn.Wordnet("test-en")
    assert set(leaves(en, pos="n")) == {
        en.synset("test-en-0005-n"),
        en.synset("test-en-0006-n"),
        en.synset("test-en-0008-n"),
    }
    assert set(leaves(en, pos="v")) == {
        en.synset("test-en-0003-v"),
        en.synset("test-en-0007-v"),
    }


@pytest.mark.usefixtures("mini_db")
def test_taxonomy_depth():
    en = wn.Wordnet("test-en")
    assert taxonomy_depth(en, pos="n") == 3
    assert taxonomy_depth(en, pos="v") == 0


@pytest.mark.usefixtures("mini_db")
def test_hypernym_paths():
    information = wn.synsets("information")[0]
    example = wn.synsets("example")[0]
    sample = wn.synsets("sample")[0]
    random_sample = wn.synsets("random sample")[0]
    assert hypernym_paths(information) == []
    assert hypernym_paths(example) == [[information]]
    assert hypernym_paths(sample) == [[example, information]]
    assert hypernym_paths(random_sample) == [[sample, example, information]]


@pytest.mark.usefixtures("mini_db")
def test_interlingual_hypernym_paths():
    información = wn.synsets("información")[0]
    ejemplo = wn.synsets("ejemplo")[0]
    inferred = wn.Synset.empty("*INFERRED*")
    muestra_aleatoria = wn.synsets("muestra aleatoria")[0]
    assert hypernym_paths(información) == []
    assert hypernym_paths(ejemplo) == [[información]]
    assert hypernym_paths(muestra_aleatoria) == [[inferred, ejemplo, información]]


@pytest.mark.usefixtures("mini_db")
def test_shortest_path():
    information = wn.synsets("information")[0]
    example = wn.synsets("example")[0]
    sample = wn.synsets("sample")[0]
    random_sample = wn.synsets("random sample")[0]
    datum = wn.synsets("datum")[0]
    exemplify = wn.synsets("exemplify")[0]
    inferred_root = wn.Synset.empty("*INFERRED*")
    assert shortest_path(information, information) == []
    assert shortest_path(information, datum) == [datum]
    assert shortest_path(information, sample) == [example, sample]
    assert shortest_path(sample, information) == [example, information]
    assert shortest_path(random_sample, datum) == [sample, example, information, datum]
    with pytest.raises(wn.Error):
        shortest_path(example, exemplify)
    assert shortest_path(example, exemplify, simulate_root=True) == [
        information,
        inferred_root,
        exemplify,
    ]


@pytest.mark.usefixtures("mini_db")
def test_min_depth():
    assert min_depth(wn.synsets("information")[0]) == 0
    assert min_depth(wn.synsets("example")[0]) == 1
    assert min_depth(wn.synsets("sample")[0]) == 2
    assert min_depth(wn.synsets("random sample")[0]) == 3


@pytest.mark.usefixtures("mini_db")
def test_max_depth():
    assert max_depth(wn.synsets("information")[0]) == 0
    assert max_depth(wn.synsets("example")[0]) == 1
    assert max_depth(wn.synsets("sample")[0]) == 2
    assert max_depth(wn.synsets("random sample")[0]) == 3



"""
Wordnet and ILI Packages and Collections
"""

from typing import Optional, Iterator, List, Tuple
from pathlib import Path
import tarfile
import tempfile
from contextlib import contextmanager
import gzip
import lzma
import shutil


import wn
from wn._types import AnyPath
from wn.constants import _WORDNET, _ILI
from wn._util import is_gzip, is_lzma
from wn import lmf
from wn import _ili


_ADDITIONAL_FILE_SUFFIXES = ("", ".txt", ".md", ".rst")


def is_package_directory(path: AnyPath) -> bool:
    """Return ``True`` if *path* appears to be a wordnet or ILI package."""
    path = Path(path).expanduser()
    return len(_package_directory_types(path)) == 1


def _package_directory_types(path: Path) -> List[Tuple[Path, str]]:
    types: List[Tuple[Path, str]] = []
    if path.is_dir():
        for p in path.iterdir():
            typ = _resource_file_type(p)
            if typ is not None:
                types.append((p, typ))
    return types


def _resource_file_type(path: Path) -> Optional[str]:
    if lmf.is_lmf(path):
        return _WORDNET
    elif _ili.is_ili(path):
        return _ILI
    return None


def is_collection_directory(path: AnyPath) -> bool:
    """Return ``True`` if *path* appears to be a wordnet collection."""
    path = Path(path).expanduser()
    return (
        path.is_dir() and len(list(filter(is_package_directory, path.iterdir()))) >= 1
    )


class _Project:
    __slots__ = ("_path",)

    def __init__(self, path: AnyPath):
        self._path: Path = Path(path).expanduser()

    def readme(self) -> Optional[Path]:
        """Return the path of the README file, or ``None`` if none exists."""
        return self._find_file(self._path / "README", _ADDITIONAL_FILE_SUFFIXES)

    def license(self) -> Optional[Path]:
        """Return the path of the license, or ``None`` if none exists."""
        return self._find_file(self._path / "LICENSE", _ADDITIONAL_FILE_SUFFIXES)

    def citation(self) -> Optional[Path]:
        """Return the path of the citation, or ``None`` if none exists."""
        return self._find_file(self._path / "citation", (".bib",))

    def _find_file(self, base: Path, suffixes: Tuple[str, ...]) -> Optional[Path]:
        for suffix in suffixes:
            base = base.with_suffix(suffix)
            if base.is_file():
                return base
        return None


class Package(_Project):
    """This class represents a wordnet or ILI package -- a directory with
    a resource file and optional metadata.

    """

    @property
    def type(self) -> Optional[str]:
        return _resource_file_type(self.resource_file())

    def resource_file(self) -> Path:
        """Return the path of the package's resource file."""
        files = _package_directory_types(self._path)
        if not files:
            raise wn.Error(f"no resource found in package: {self._path!s}")
        elif len(files) > 1:
            raise wn.Error(f"multiple resource found in package: {self._path!s}")
        return files[0][0]


class _ResourceOnlyPackage(Package):
    def resource_file(self) -> Path:
        return self._path

    def readme(self):
        return None

    def license(self):
        return None

    def citation(self):
        return None


class Collection(_Project):
    """This class represents a wordnet or ILI collection -- a directory
    with one or more wordnet/ILI packages and optional metadata.

    """

    def packages(self) -> List[Package]:
        """Return the list of packages in the collection."""
        return [
            Package(path) for path in self._path.iterdir() if is_package_directory(path)
        ]


def iterpackages(path: AnyPath) -> Iterator[Package]:
    """Yield any wordnet or ILI packages found at *path*.

    The *path* argument can point to one of the following:
      - a lexical resource file or ILI file
      - a wordnet package directory
      - a wordnet collection directory
      - a tar archive containing one of the above
      - a compressed (gzip or lzma) resource file or tar archive
    """
    path = Path(path).expanduser()

    if path.is_dir():
        if is_package_directory(path):
            yield Package(path)

        elif is_collection_directory(path):
            yield from Collection(path).packages()

        else:
            raise wn.Error(
                f"does not appear to be a valid package or collection: {path!s}"
            )

    elif tarfile.is_tarfile(path):
        with tarfile.open(path) as tar:
            _check_tar(tar)
            with tempfile.TemporaryDirectory() as tmpdir:
                tar.extractall(path=tmpdir)
                contents = list(Path(tmpdir).iterdir())
                if len(contents) != 1:
                    raise wn.Error(
                        "archive may only have one resource, package, or collection"
                    )
                yield from iterpackages(contents[0])

    else:
        decompressed: Path
        with _get_decompressed(path) as decompressed:
            if lmf.is_lmf(decompressed) or _ili.is_ili(decompressed):
                yield _ResourceOnlyPackage(decompressed)
            else:
                raise wn.Error(f"not a valid lexical resource: {path!s}")


@contextmanager
def _get_decompressed(source: Path) -> Iterator[Path]:
    gzipped = is_gzip(source)
    xzipped = is_lzma(source)
    if not (gzipped or xzipped):
        yield source
    else:
        tmp = tempfile.NamedTemporaryFile(suffix=".xml", delete=False)
        path = Path(tmp.name)
        try:
            if gzipped:
                with gzip.open(source, "rb") as gzip_src:
                    shutil.copyfileobj(gzip_src, tmp)
            else:  # xzipped
                with lzma.open(source, "rb") as lzma_src:
                    shutil.copyfileobj(lzma_src, tmp)

            tmp.close()  # Windows cannot reliably reopen until it's closed

            yield path

        except (OSError, EOFError, lzma.LZMAError) as exc:
            raise wn.Error(f"could not decompress file: {source}") from exc

        finally:
            path.unlink()


def _check_tar(tar: tarfile.TarFile) -> None:
    """Check the tarfile to avoid potential security issues.

    Currently collections and packages have the following constraints:
    - Only regular files or directories
    - No paths starting with '/' or containing '..'
    """
    for info in tar.getmembers():
        if not (info.isfile() or info.isdir()):
            raise wn.Error(
                f"tarfile member is not a regular file or directory: {info.name}"
            )
        if info.name.startswith("/") or ".." in info.name:
            raise wn.Error(
                f"tarfile member paths may not be absolute or contain ..: {info.name}"
            )



import pytest

import wn
from wn import morphy


def test_morphy_uninitialized():
    # An unintialized Morphy isn't very bright, but it starts up
    # fast. It relies on the database to filter bad items.
    m = morphy.Morphy()
    assert m("example", "n") == {"n": {"example"}}
    assert m("examples", "n") == {"n": {"examples", "example"}}
    assert m("examples", "v") == {"v": {"examples", "example", "exampl"}}
    assert m("exemplifying", "n") == {"n": {"exemplifying"}}
    assert m("exemplifying", "v") == {"v": {"exemplifying", "exemplify", "exemplifye"}}
    assert m("data", "n") == {"n": {"data"}}
    assert m("datums", "n") == {"n": {"datums", "datum"}}  # expected false positive
    assert m("examples", None) == {
        None: {"examples"},
        "n": {"example"},
        "v": {"example", "exampl"},
    }
    assert m("exemplifying", None) == {
        None: {"exemplifying"},
        "v": {"exemplify", "exemplifye"},
    }
    assert m("data", None) == {None: {"data"}}


@pytest.mark.usefixtures("mini_db")
def test_morphy_initialized():
    w = wn.Wordnet("test-en:1")
    m = morphy.Morphy(wordnet=w)
    assert m("example", "n") == {"n": {"example"}}
    assert m("examples", "n") == {"n": {"example"}}
    assert m("examples", "v") == {}
    assert m("exemplifying", "n") == {}
    assert m("exemplifying", "v") == {"v": {"exemplify"}}
    assert m("data", "n") == {"n": {"datum"}}
    assert m("datums", "n") == {"n": {"datum"}}  # expected false positive
    assert m("examples", None) == {"n": {"example"}}
    assert m("exemplifying", None) == {"v": {"exemplify"}}
    assert m("data", None) == {"n": {"datum"}}


@pytest.mark.usefixtures("mini_db")
def test_issue_154():
    # https://github.com/goodmami/wn/issues/154
    w = wn.Wordnet("test-en:1")
    assert w.words("exemplifies") == [w.word("test-en-exemplify-v")]
    assert w.words("samples") == []
    w = wn.Wordnet("test-en:1", lemmatizer=morphy.Morphy())
    assert w.words("exemplifies") == [w.word("test-en-exemplify-v")]
    assert w.words("samples") == [w.word("test-en-sample-n")]



from wn._core import Word, Synset


# Word-based Metrics


def ambiguity(word: Word) -> int:
    return len(word.synsets())


def average_ambiguity(synset: Synset) -> float:
    words = synset.words()
    return sum(len(word.synsets()) for word in words) / len(words)



"""Functions for working with hypernym/hyponym taxonomies."""

from typing import Optional, Tuple, List, Set, Dict, TYPE_CHECKING

import wn
from wn.constants import ADJ, ADJ_SAT
from wn._util import flatten
from wn import _core

if TYPE_CHECKING:
    from wn._core import Wordnet, Synset


_FAKE_ROOT = "*ROOT*"


def roots(wordnet: "Wordnet", pos: Optional[str] = None) -> List["Synset"]:
    """Return the list of root synsets in *wordnet*.

    Arguments:

        wordnet: The wordnet from which root synsets are found.

        pos: If given, only return synsets with the specified part of
            speech.

    Example:

        >>> import wn, wn.taxonomy
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> len(wn.taxonomy.roots(ewn, pos='v'))
        573


    """
    return [ss for ss in _synsets_for_pos(wordnet, pos) if not ss.hypernyms()]


def leaves(wordnet: "Wordnet", pos: Optional[str] = None) -> List["Synset"]:
    """Return the list of leaf synsets in *wordnet*.

    Arguments:

        wordnet: The wordnet from which leaf synsets are found.

        pos: If given, only return synsets with the specified part of
            speech.

    Example:

        >>> import wn, wn.taxonomy
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> len(wn.taxonomy.leaves(ewn, pos='v'))
        10525

    """
    return [ss for ss in _synsets_for_pos(wordnet, pos) if not ss.hyponyms()]


def taxonomy_depth(wordnet: "Wordnet", pos: str) -> int:
    """Return the list of leaf synsets in *wordnet*.

    Arguments:

        wordnet: The wordnet for which the taxonomy depth will be
            calculated.

        pos: The part of speech for which the taxonomy depth will be
            calculated.

    Example:

        >>> import wn, wn.taxonomy
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> wn.taxonomy.taxonomy_depth(ewn, 'n')
        19

    """
    seen: Set["Synset"] = set()
    depth = 0
    for ss in _synsets_for_pos(wordnet, pos):
        if all(hyp in seen for hyp in ss.hypernyms()):
            continue
        paths = ss.hypernym_paths()
        if paths:
            depth = max(depth, max(len(path) for path in paths))
            seen.update(hyp for path in paths for hyp in path)
    return depth


def _synsets_for_pos(wordnet: "Wordnet", pos: Optional[str]) -> List["Synset"]:
    """Get the list of synsets for a part of speech. If *pos* is 'a' or
    's', also include those for the other.

    """
    synsets = wordnet.synsets(pos=pos)
    if pos == ADJ:
        synsets.extend(wordnet.synsets(pos=ADJ_SAT))
    elif pos == ADJ_SAT:
        synsets.extend(wordnet.synsets(pos=ADJ))
    return synsets


def _hypernym_paths(
    synset: "Synset", simulate_root: bool, include_self: bool
) -> List[List["Synset"]]:
    paths = list(synset.relation_paths("hypernym", "instance_hypernym"))
    if include_self:
        paths = [[synset] + path for path in paths] or [[synset]]
    if simulate_root and synset.id != _FAKE_ROOT:
        root = _core.Synset.empty(
            id=_FAKE_ROOT, _lexid=synset._lexid, _wordnet=synset._wordnet
        )
        paths = [path + [root] for path in paths] or [[root]]
    return paths


def hypernym_paths(
    synset: "Synset", simulate_root: bool = False
) -> List[List["Synset"]]:
    """Return the list of hypernym paths to a root synset.

    Arguments:

        synset: The starting synset for paths to a root.

        simulate_root: If :python:`True`, find the path to a simulated
            root node.

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = wn.synsets('dog', pos='n')[0]
        >>> for path in wn.taxonomy.hypernym_paths(dog):
        ...     for i, ss in enumerate(path):
        ...         print(' ' * i, ss, ss.lemmas()[0])
        ...
         Synset('pwn-02083346-n') canine
          Synset('pwn-02075296-n') carnivore
           Synset('pwn-01886756-n') eutherian mammal
            Synset('pwn-01861778-n') mammalian
             Synset('pwn-01471682-n') craniate
              Synset('pwn-01466257-n') chordate
               Synset('pwn-00015388-n') animal
                Synset('pwn-00004475-n') organism
                 Synset('pwn-00004258-n') animate thing
                  Synset('pwn-00003553-n') unit
                   Synset('pwn-00002684-n') object
                    Synset('pwn-00001930-n') physical entity
                     Synset('pwn-00001740-n') entity
         Synset('pwn-01317541-n') domesticated animal
          Synset('pwn-00015388-n') animal
           Synset('pwn-00004475-n') organism
            Synset('pwn-00004258-n') animate thing
             Synset('pwn-00003553-n') unit
              Synset('pwn-00002684-n') object
               Synset('pwn-00001930-n') physical entity
                Synset('pwn-00001740-n') entity

    """
    return _hypernym_paths(synset, simulate_root, False)


def min_depth(synset: "Synset", simulate_root: bool = False) -> int:
    """Return the minimum taxonomy depth of the synset.

    Arguments:

        synset: The starting synset for paths to a root.

        simulate_root: If :python:`True`, find the depth to a
            simulated root node.

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = wn.synsets('dog', pos='n')[0]
        >>> wn.taxonomy.min_depth(dog)
        8

    """
    return min(
        (len(path) for path in synset.hypernym_paths(simulate_root=simulate_root)),
        default=0,
    )


def max_depth(synset: "Synset", simulate_root: bool = False) -> int:
    """Return the maximum taxonomy depth of the synset.

    Arguments:

        synset: The starting synset for paths to a root.

        simulate_root: If :python:`True`, find the depth to a
            simulated root node.

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = wn.synsets('dog', pos='n')[0]
        >>> wn.taxonomy.max_depth(dog)
        13

    """
    return max(
        (len(path) for path in synset.hypernym_paths(simulate_root=simulate_root)),
        default=0,
    )


def _shortest_hyp_paths(
    synset: "Synset", other: "Synset", simulate_root: bool
) -> Dict[Tuple["Synset", int], List["Synset"]]:
    if synset == other:
        return {(synset, 0): []}

    from_self = _hypernym_paths(synset, simulate_root, True)
    from_other = _hypernym_paths(other, simulate_root, True)
    common = set(flatten(from_self)).intersection(flatten(from_other))

    if not common:
        return {}

    # Compute depths of common hypernyms from their distances.
    # Doing this now avoid more expensive lookups later.
    depths: Dict["Synset", int] = {}
    # subpaths accumulates paths to common hypernyms from both sides
    subpaths: Dict["Synset", Tuple[List[List["Synset"]], List[List["Synset"]]]]
    subpaths = {ss: ([], []) for ss in common}
    for which, paths in (0, from_self), (1, from_other):
        for path in paths:
            for dist, ss in enumerate(path):
                if ss in common:
                    # synset or other subpath to ss (not including ss)
                    subpaths[ss][which].append(path[: dist + 1])
                    # keep maximum depth
                    depth = len(path) - dist - 1
                    if ss not in depths or depths[ss] < depth:
                        depths[ss] = depth

    shortest: Dict[Tuple["Synset", int], List["Synset"]] = {}
    for ss in common:
        from_self_subpaths, from_other_subpaths = subpaths[ss]
        shortest_from_self = min(from_self_subpaths, key=len)
        # for the other path, we need to reverse it and remove the pivot synset
        shortest_from_other = min(from_other_subpaths, key=len)[-2::-1]
        shortest[(ss, depths[ss])] = shortest_from_self + shortest_from_other

    return shortest


def shortest_path(
    synset: "Synset", other: "Synset", simulate_root: bool = False
) -> List["Synset"]:
    """Return the shortest path from *synset* to the *other* synset.

    Arguments:
        other: endpoint synset of the path
        simulate_root: if :python:`True`, ensure any two synsets
          are always connected by positing a fake root node

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = ewn.synsets('dog', pos='n')[0]
        >>> squirrel = ewn.synsets('squirrel', pos='n')[0]
        >>> for ss in wn.taxonomy.shortest_path(dog, squirrel):
        ...     print(ss.lemmas())
        ...
        ['canine', 'canid']
        ['carnivore']
        ['eutherian mammal', 'placental', 'placental mammal', 'eutherian']
        ['rodent', 'gnawer']
        ['squirrel']

    """
    pathmap = _shortest_hyp_paths(synset, other, simulate_root)
    key = min(pathmap, key=lambda key: len(pathmap[key]), default=None)
    if key is None:
        raise wn.Error(f"no path between {synset!r} and {other!r}")
    return pathmap[key][1:]


def common_hypernyms(
    synset: "Synset", other: "Synset", simulate_root: bool = False
) -> List["Synset"]:
    """Return the common hypernyms for the current and *other* synsets.

    Arguments:
        other: synset that is a hyponym of any shared hypernyms
        simulate_root: if :python:`True`, ensure any two synsets
          always share a hypernym by positing a fake root node

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = ewn.synsets('dog', pos='n')[0]
        >>> squirrel = ewn.synsets('squirrel', pos='n')[0]
        >>> for ss in wn.taxonomy.common_hypernyms(dog, squirrel):
        ...     print(ss.lemmas())
        ...
        ['entity']
        ['physical entity']
        ['object', 'physical object']
        ['unit', 'whole']
        ['animate thing', 'living thing']
        ['organism', 'being']
        ['fauna', 'beast', 'animate being', 'brute', 'creature', 'animal']
        ['chordate']
        ['craniate', 'vertebrate']
        ['mammalian', 'mammal']
        ['eutherian mammal', 'placental', 'placental mammal', 'eutherian']

    """
    from_self = _hypernym_paths(synset, simulate_root, True)
    from_other = _hypernym_paths(other, simulate_root, True)
    common = set(flatten(from_self)).intersection(flatten(from_other))
    return sorted(common)


def lowest_common_hypernyms(
    synset: "Synset", other: "Synset", simulate_root: bool = False
) -> List["Synset"]:
    """Return the common hypernyms furthest from the root.

    Arguments:
        other: synset that is a hyponym of any shared hypernyms
        simulate_root: if :python:`True`, ensure any two synsets
          always share a hypernym by positing a fake root node

    Example:

        >>> import wn, wn.taxonomy
        >>> dog = ewn.synsets('dog', pos='n')[0]
        >>> squirrel = ewn.synsets('squirrel', pos='n')[0]
        >>> len(wn.taxonomy.lowest_common_hypernyms(dog, squirrel))
        1
        >>> wn.taxonomy.lowest_common_hypernyms(dog, squirrel)[0].lemmas()
        ['eutherian mammal', 'placental', 'placental mammal', 'eutherian']

    """
    pathmap = _shortest_hyp_paths(synset, other, simulate_root)
    # keys of pathmap are (synset, depth_of_synset)
    max_depth: int = max([depth for _, depth in pathmap], default=-1)
    if max_depth == -1:
        return []
    else:
        return [ss for ss, d in pathmap if d == max_depth]



from typing import Iterator, Dict
from pathlib import Path

from wn._types import AnyPath


def is_ili(source: AnyPath) -> bool:
    """Return True if *source* is an ILI tab-separated-value file.

    This only checks that the first column, split by tabs, of the
    first line is 'ili' or 'ILI'. It does not check if each line has
    the correct number of columns.

    """
    source = Path(source).expanduser()
    if source.is_file():
        try:
            with source.open("rb") as fh:
                return next(fh).split(b"\t")[0] in (b"ili", b"ILI")
        except (StopIteration, IndexError):
            pass
    return False


def load(source: AnyPath) -> Iterator[Dict[str, str]]:
    """Load an interlingual index file.

    Args:
        source: path to an ILI file
    """
    source = Path(source).expanduser()
    with source.open(encoding="utf-8") as fh:
        header = next(fh).rstrip("\r\n")
        fields = tuple(map(str.lower, header.split("\t")))
        for line in fh:
            yield dict(zip(fields, line.rstrip("\r\n").split("\t")))



"""Synset similarity metrics."""

from typing import List
import math

import wn
from wn.constants import ADJ, ADJ_SAT
from wn._core import Synset
from wn.ic import Freq, information_content


def path(synset1: Synset, synset2: Synset, simulate_root: bool = False) -> float:
    """Return the Path similarity of *synset1* and *synset2*.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        simulate_root: When :python:`True`, a fake root node connects
            all other roots; default: :python:`False`.

    Example:
        >>> import wn
        >>> from wn.similarity import path
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> spatula = ewn.synsets('spatula')[0]
        >>> path(spatula, ewn.synsets('pancake')[0])
        0.058823529411764705
        >>> path(spatula, ewn.synsets('utensil')[0])
        0.2
        >>> path(spatula, spatula)
        1.0
        >>> flip = ewn.synsets('flip', pos='v')[0]
        >>> turn_over = ewn.synsets('turn over', pos='v')[0]
        >>> path(flip, turn_over)
        0.0
        >>> path(flip, turn_over, simulate_root=True)
        0.16666666666666666

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    try:
        path = synset1.shortest_path(synset2, simulate_root=simulate_root)
    except wn.Error:
        distance = float("inf")
    else:
        distance = len(path)
    return 1 / (distance + 1)


def wup(synset1: Synset, synset2: Synset, simulate_root=False) -> float:
    """Return the Wu-Palmer similarity of *synset1* and *synset2*.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        simulate_root: When :python:`True`, a fake root node connects
            all other roots; default: :python:`False`.

    Raises:
        wn.Error: When no path connects the *synset1* and *synset2*.

    Example:
        >>> import wn
        >>> from wn.similarity import wup
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> spatula = ewn.synsets('spatula')[0]
        >>> wup(spatula, ewn.synsets('pancake')[0])
        0.2
        >>> wup(spatula, ewn.synsets('utensil')[0])
        0.8
        >>> wup(spatula, spatula)
        1.0
        >>> flip = ewn.synsets('flip', pos='v')[0]
        >>> turn_over = ewn.synsets('turn over', pos='v')[0]
        >>> wup(flip, turn_over, simulate_root=True)
        0.2857142857142857

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    lcs_list = _least_common_subsumers(synset1, synset2, simulate_root)
    lcs = lcs_list[0]
    i = len(synset1.shortest_path(lcs, simulate_root=simulate_root))
    j = len(synset2.shortest_path(lcs, simulate_root=simulate_root))
    k = lcs.max_depth() + 1
    return (2 * k) / (i + j + 2 * k)


def lch(
    synset1: Synset, synset2: Synset, max_depth: int, simulate_root: bool = False
) -> float:
    """Return the Leacock-Chodorow similarity between *synset1* and *synset2*.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        max_depth: The taxonomy depth (see :func:`wn.taxonomy.taxonomy_depth`)
        simulate_root: When :python:`True`, a fake root node connects
            all other roots; default: :python:`False`.

    Example:
        >>> import wn, wn.taxonomy
        >>> from wn.similarity import lch
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> n_depth = wn.taxonomy.taxonomy_depth(ewn, 'n')
        >>> spatula = ewn.synsets('spatula')[0]
        >>> lch(spatula, ewn.synsets('pancake')[0], n_depth)
        0.8043728156701697
        >>> lch(spatula, ewn.synsets('utensil')[0], n_depth)
        2.0281482472922856
        >>> lch(spatula, spatula, n_depth)
        3.6375861597263857
        >>> v_depth = taxonomy.taxonomy_depth(ewn, 'v')
        >>> flip = ewn.synsets('flip', pos='v')[0]
        >>> turn_over = ewn.synsets('turn over', pos='v')[0]
        >>> lch(flip, turn_over, v_depth, simulate_root=True)
        1.3862943611198906

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    distance = len(synset1.shortest_path(synset2, simulate_root=simulate_root))
    if max_depth <= 0:
        raise wn.Error("max_depth must be greater than 0")
    return -math.log((distance + 1) / (2 * max_depth))


def res(synset1: Synset, synset2: Synset, ic: Freq) -> float:
    """Return the Resnik similarity between *synset1* and *synset2*.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        ic: Information Content weights.

    Example:
        >>> import wn, wn.ic, wn.taxonomy
        >>> from wn.similarity import res
        >>> pwn = wn.Wordnet('pwn:3.0')
        >>> ic = wn.ic.load('~/nltk_data/corpora/wordnet_ic/ic-brown.dat', pwn)
        >>> spatula = pwn.synsets('spatula')[0]
        >>> res(spatula, pwn.synsets('pancake')[0], ic)
        0.8017591149538994
        >>> res(spatula, pwn.synsets('utensil')[0], ic)
        5.87738923441087

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    lcs = _most_informative_lcs(synset1, synset2, ic)
    return information_content(lcs, ic)


def jcn(synset1: Synset, synset2: Synset, ic: Freq) -> float:
    """Return the Jiang-Conrath similarity of two synsets.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        ic: Information Content weights.

    Example:
        >>> import wn, wn.ic, wn.taxonomy
        >>> from wn.similarity import jcn
        >>> pwn = wn.Wordnet('pwn:3.0')
        >>> ic = wn.ic.load('~/nltk_data/corpora/wordnet_ic/ic-brown.dat', pwn)
        >>> spatula = pwn.synsets('spatula')[0]
        >>> jcn(spatula, pwn.synsets('pancake')[0], ic)
        0.04061799236354239
        >>> jcn(spatula, pwn.synsets('utensil')[0], ic)
        0.10794048564613007

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    ic1 = information_content(synset1, ic)
    ic2 = information_content(synset2, ic)
    lcs = _most_informative_lcs(synset1, synset2, ic)
    ic_lcs = information_content(lcs, ic)
    if ic1 == ic2 == ic_lcs == 0:
        return 0
    elif ic1 + ic2 == 2 * ic_lcs:
        return float("inf")
    else:
        return 1 / (ic1 + ic2 - 2 * ic_lcs)


def lin(synset1: Synset, synset2: Synset, ic: Freq) -> float:
    """Return the Lin similarity of two synsets.

    Arguments:
        synset1: The first synset to compare.
        synset2: The second synset to compare.
        ic: Information Content weights.

    Example:
        >>> import wn, wn.ic, wn.taxonomy
        >>> from wn.similarity import lin
        >>> pwn = wn.Wordnet('pwn:3.0')
        >>> ic = wn.ic.load('~/nltk_data/corpora/wordnet_ic/ic-brown.dat', pwn)
        >>> spatula = pwn.synsets('spatula')[0]
        >>> lin(spatula, pwn.synsets('pancake')[0], ic)
        0.061148956278604116
        >>> lin(spatula, pwn.synsets('utensil')[0], ic)
        0.5592415686750427

    """
    _check_if_pos_compatible(synset1.pos, synset2.pos)
    lcs = _most_informative_lcs(synset1, synset2, ic)
    ic1 = information_content(synset1, ic)
    ic2 = information_content(synset2, ic)
    if ic1 == 0 or ic2 == 0:
        return 0.0
    return 2 * information_content(lcs, ic) / (ic1 + ic2)


# Helper functions


def _least_common_subsumers(
    synset1: Synset, synset2: Synset, simulate_root: bool
) -> List[Synset]:
    lcs = synset1.lowest_common_hypernyms(synset2, simulate_root=simulate_root)
    if not lcs:
        raise wn.Error(f"no common hypernyms for {synset1!r} and {synset2!r}")
    return lcs


def _most_informative_lcs(synset1: Synset, synset2: Synset, ic: Freq) -> Synset:
    pos_ic = ic[synset1.pos]
    lcs = _least_common_subsumers(synset1, synset2, False)
    return max(lcs, key=lambda ss: pos_ic[ss.id])


def _check_if_pos_compatible(pos1: str, pos2: str) -> None:
    _pos1 = ADJ if pos1 == ADJ_SAT else pos1
    _pos2 = ADJ if pos2 == ADJ_SAT else pos2
    if _pos1 != _pos2:
        raise wn.Error("synsets must have the same part of speech")



from typing import Optional, Type, List, Sequence, Tuple
from pathlib import Path
import logging

import requests

import wn
from wn._util import is_url
from wn.util import ProgressHandler, ProgressBar
from wn._add import add as add_to_db
from wn import config


CHUNK_SIZE = 8 * 1024  # how many KB to read at a time
TIMEOUT = 10  # number of seconds to wait for a server response


logger = logging.getLogger("wn")


def download(
    project_or_url: str,
    add: bool = True,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> Path:
    """Download the resource specified by *project_or_url*.

    First the URL of the resource is determined and then, depending on
    the parameters, the resource is downloaded and added to the
    database.  The function then returns the path of the cached file.

    If *project_or_url* starts with `'http://'` or `'https://'`, then
    it is taken to be the URL for the resource. Otherwise,
    *project_or_url* is taken as a :ref:`project specifier
    <lexicon-specifiers>` and the URL is taken from a matching entry
    in Wn's project index. If no project matches the specifier,
    :exc:`wn.Error` is raised.

    If the URL has been downloaded and cached before, the cached file
    is used. Otherwise the URL is retrieved and stored in the cache.

    If the *add* paramter is ``True`` (default), the downloaded
    resource is added to the database.

    >>> wn.download('ewn:2020')
    Added ewn:2020 (English WordNet)

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.

    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message="Download", unit=" bytes")

    cache_path, urls = _get_cache_path_and_urls(project_or_url)

    try:
        if cache_path and cache_path.exists():
            progress.flash(f"Cached file found: {cache_path!s}")
            path = cache_path
        elif urls:
            path = _download(urls, progress)
        else:
            raise wn.Error("no urls to download")
    finally:
        progress.close()

    if add:
        try:
            add_to_db(path, progress_handler=progress_handler)
        except wn.Error as exc:
            raise wn.Error(
                f"could not add downloaded file: {path}\n  You might try "
                "deleting the cached file and trying the download again."
            ) from exc

    return path


def _get_cache_path_and_urls(project_or_url: str) -> Tuple[Optional[Path], List[str]]:
    if is_url(project_or_url):
        return config.get_cache_path(project_or_url), [project_or_url]
    else:
        info = config.get_project_info(project_or_url)
        return info.get("cache"), info["resource_urls"]


def _download(urls: Sequence[str], progress: ProgressHandler) -> Path:
    try:
        for i, url in enumerate(urls, 1):
            path = config.get_cache_path(url)
            logger.info("download url: %s", url)
            logger.info("download cache path: %s", path)
            try:
                with open(path, "wb") as f:
                    progress.set(status="Requesting", count=0)
                    with requests.get(url, stream=True, timeout=TIMEOUT) as response:
                        response.raise_for_status()
                        size = int(response.headers.get("Content-Length", 0))
                        progress.set(total=size, status="Receiving")
                        for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                            if chunk:
                                f.write(chunk)
                            progress.update(len(chunk))
                        progress.set(status="Complete")
            except requests.exceptions.RequestException as exc:
                _unlink_if_exists(path)
                count = progress.kwargs["count"]
                if i == len(urls):
                    raise wn.Error(f"download failed at {count} bytes") from exc
                else:
                    logger.info("download failed at %d bytes; trying next url", count)
            else:
                break  # success

    except KeyboardInterrupt as exc:
        _unlink_if_exists(path)
        count = progress.kwargs["count"]
        raise wn.Error(f"download cancelled at {count} bytes") from exc
    except Exception:
        _unlink_if_exists(path)
        raise

    return path


def _unlink_if_exists(path: Path) -> None:
    """Unlink *path* if it exists.

    From Python 3.8, this function is unnecessary. Just use:
        path.unlink(exist_ok=True)
    """
    if path.exists():
        path.unlink()



"""
Adding and removing lexicons to/from the database.
"""

from typing import (
    Union,
    Optional,
    TypeVar,
    Type,
    Tuple,
    List,
    Dict,
    Set,
    Iterator,
    Iterable,
    Sequence,
    cast,
)
from pathlib import Path
from itertools import islice
import sqlite3
import logging

import wn
from wn._types import AnyPath
from wn.constants import _WORDNET, _ILI
from wn._db import connect
from wn._queries import find_lexicons, get_lexicon_extensions, get_lexicon
from wn._util import normalize_form
from wn.util import ProgressHandler, ProgressBar
from wn.project import iterpackages
from wn import lmf
from wn import _ili


logger = logging.getLogger("wn")


BATCH_SIZE = 1000
DEFAULT_MEMBER_RANK = 127  # synset member rank when not specified by 'members'

ENTRY_QUERY = """
    SELECT e.rowid
      FROM entries AS e
     WHERE e.id = ?
       AND e.lexicon_rowid = ?
"""
# forms don't have reliable ids, so also consider rank; this depends
# on each form having a unique rank, and this doesn't work for lexicon
# extensions
FORM_QUERY = """
    SELECT f.rowid
      FROM forms AS f
      JOIN entries AS e ON f.entry_rowid = e.rowid
     WHERE e.id = ?
       AND e.lexicon_rowid = ?
       AND (f.id = ? OR f.rank = ?)
"""
SENSE_QUERY = """
    SELECT s.rowid
      FROM senses AS s
     WHERE s.id = ?
       AND s.lexicon_rowid = ?
"""
SYNSET_QUERY = """
    SELECT ss.rowid
      FROM synsets AS ss
     WHERE ss.id = ?
       AND ss.lexicon_rowid = ?
"""
RELTYPE_QUERY = """
    SELECT rt.rowid
      FROM relation_types AS rt
     WHERE rt.type = ?
"""
ILISTAT_QUERY = """
    SELECT ist.rowid
      FROM ili_statuses AS ist
     WHERE ist.status = ?
"""
LEXFILE_QUERY = """
    SELECT lf.rowid
      FROM lexfiles AS lf
     WHERE lf.name = ?
"""

_AnyLexicon = Union[lmf.Lexicon, lmf.LexiconExtension]
_AnyEntry = Union[lmf.LexicalEntry, lmf.ExternalLexicalEntry]
_AnyLemma = Union[lmf.Lemma, lmf.ExternalLemma]
_AnyForm = Union[lmf.Form, lmf.ExternalForm]
_AnySense = Union[lmf.Sense, lmf.ExternalSense]
_AnySynset = Union[lmf.Synset, lmf.ExternalSynset]


def add(
    source: AnyPath,
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> None:
    """Add the LMF file at *source* to the database.

    The file at *source* may be gzip-compressed or plain text XML.

    >>> wn.add('english-wordnet-2020.xml')
    Added ewn:2020 (English WordNet)

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.

    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message="Database")

    logger.info("adding project to database")
    logger.info("  database: %s", wn.config.database_path)
    logger.info("  project file: %s", source)

    try:
        for package in iterpackages(source):
            if package.type == _WORDNET:
                _add_lmf(package.resource_file(), progress, progress_handler)
            elif package.type == _ILI:
                _add_ili(package.resource_file(), progress)
            else:
                raise wn.Error(f"unknown package type: {package.type}")
    finally:
        progress.close()


def _add_lmf(
    source: Path,
    progress: ProgressHandler,
    progress_handler: Type[ProgressHandler],
) -> None:
    with connect() as conn:
        cur = conn.cursor()
        # these two settings increase the risk of database corruption
        # if the system crashes during a write, but they should also
        # make inserts much faster
        cur.execute("PRAGMA synchronous = OFF")
        cur.execute("PRAGMA journal_mode = MEMORY")

        # abort if any lexicon in *source* is already added
        progress.flash(f"Checking {source!s}")
        all_infos = _precheck(source, cur)

        if not all_infos:
            progress.flash(f"{source}: No lexicons found")
            return
        for info in all_infos:
            skip = info.get("skip", "")
            if skip:
                id, ver, lbl = info["id"], info["version"], info["label"]
                progress.flash(f"Skipping {id}:{ver} ({lbl}); {skip}\n")
        if all("skip" in info for info in all_infos):
            return

        # all clear, try to add them
        progress.flash(f"Reading {source!s}")
        resource = lmf.load(source, progress_handler)

        for lexicon, info in zip(resource["lexicons"], all_infos):
            if "skip" in info:
                continue

            progress.flash("Updating lookup tables")
            _update_lookup_tables(lexicon, cur)

            progress.set(count=0, total=_sum_counts(lexicon))
            synsets: Sequence[_AnySynset] = _synsets(lexicon)
            entries: Sequence[_AnyEntry] = _entries(lexicon)
            synbhrs: Sequence[lmf.SyntacticBehaviour] = _collect_frames(lexicon)

            lexid, extid = _insert_lexicon(lexicon, info, cur, progress)

            lexidmap = _build_lexid_map(lexicon, lexid, extid)

            _insert_synsets(synsets, lexid, cur, progress)
            _insert_entries(entries, lexid, cur, progress)
            _insert_forms(entries, lexid, lexidmap, cur, progress)
            _insert_pronunciations(entries, lexid, lexidmap, cur, progress)
            _insert_tags(entries, lexid, lexidmap, cur, progress)
            _insert_senses(entries, synsets, lexid, lexidmap, cur, progress)
            _insert_adjpositions(entries, lexid, lexidmap, cur, progress)
            _insert_counts(entries, lexid, lexidmap, cur, progress)
            _insert_syntactic_behaviours(synbhrs, lexid, lexidmap, cur, progress)

            _insert_synset_relations(synsets, lexid, lexidmap, cur, progress)
            _insert_sense_relations(lexicon, lexid, lexidmap, cur, progress)

            _insert_synset_definitions(synsets, lexid, lexidmap, cur, progress)
            _insert_examples(
                [sense for e in entries for sense in _senses(e)],
                lexid,
                lexidmap,
                "sense_examples",
                cur,
                progress,
            )
            _insert_examples(synsets, lexid, lexidmap, "synset_examples", cur, progress)

            progress.set(status="")  # clear type string
            progress.flash(
                f"Added {lexicon['id']}:{lexicon['version']} ({lexicon['label']})\n"
            )


def _precheck(source: Path, cur: sqlite3.Cursor) -> List[Dict]:
    lexqry = "SELECT * FROM lexicons WHERE id = :id AND version = :version"
    infos = lmf.scan_lexicons(source)
    for info in infos:
        base = info.get("extends")
        if cur.execute(lexqry, info).fetchone():
            info["skip"] = "already added"
        if base and cur.execute(lexqry, base).fetchone() is None:
            id_, ver = base["id"], base["version"]
            info["skip"] = f"base lexicon ({id_}:{ver}) not available"
    return infos


def _sum_counts(lex: _AnyLexicon) -> int:
    ents = _entries(lex)
    locs = _local_entries(ents)
    lems = [e["lemma"] for e in locs if e.get("lemma")]
    frms = [f for e in ents for f in _forms(e)]
    sens = [s for e in ents for s in _senses(e)]
    syns = _synsets(lex)
    return sum(
        [
            # lexical entries
            len(ents),
            len(lems),
            sum(len(lem.get("pronunciations", [])) for lem in lems),
            sum(len(lem.get("tags", [])) for lem in lems),
            len(frms),
            sum(len(frm.get("pronunciations", [])) for frm in frms),
            sum(len(frm.get("tags", [])) for frm in frms),
            # senses
            len(sens),
            sum(len(sen.get("relations", [])) for sen in sens),
            sum(len(sen.get("examples", [])) for sen in sens),
            sum(len(sen.get("counts", [])) for sen in sens),
            # synsets
            len(syns),
            sum(len(syn.get("definitions", [])) for syn in syns),
            sum(len(syn.get("relations", [])) for syn in syns),
            sum(len(syn.get("examples", [])) for syn in syns),
            # syntactic behaviours
            sum(len(ent.get("frames", [])) for ent in locs),
            len(lex.get("frames", [])),
        ]
    )


def _update_lookup_tables(lexicon: _AnyLexicon, cur: sqlite3.Cursor) -> None:
    reltypes = set(
        rel["relType"] for ss in _synsets(lexicon) for rel in ss.get("relations", [])
    )
    reltypes.update(
        rel["relType"]
        for e in _entries(lexicon)
        for s in _senses(e)
        for rel in s.get("relations", [])
    )
    cur.executemany(
        "INSERT OR IGNORE INTO relation_types VALUES (null,?)",
        [(rt,) for rt in sorted(reltypes)],
    )
    lexfiles: Set[str] = {
        ss.get("lexfile", "")
        for ss in _local_synsets(_synsets(lexicon))
        if ss.get("lexfile")
    }
    cur.executemany(
        "INSERT OR IGNORE INTO lexfiles VALUES (null,?)",
        [(lf,) for lf in sorted(lexfiles)],
    )


def _insert_lexicon(
    lexicon: _AnyLexicon, info: dict, cur: sqlite3.Cursor, progress: ProgressHandler
) -> Tuple[int, int]:
    progress.set(status="Lexicon Info")
    cur.execute(
        "INSERT INTO lexicons VALUES (null,?,?,?,?,?,?,?,?,?,?,?)",
        (
            lexicon["id"],
            lexicon["label"],
            lexicon["language"],
            lexicon["email"],
            lexicon["license"],
            lexicon["version"],
            lexicon.get("url"),
            lexicon.get("citation"),
            lexicon.get("logo"),
            lexicon.get("meta"),
            False,
        ),
    )
    lexid = cur.lastrowid

    if not isinstance(lexid, int):
        raise wn.Error("failed to insert lexicon")

    query = """
        UPDATE lexicon_dependencies
           SET provider_rowid = ?
         WHERE provider_id = ? AND provider_version = ?
    """
    cur.execute(query, (lexid, lexicon["id"], lexicon["version"]))

    query = """
        INSERT INTO {table}
        VALUES (:lid,
                :id,
                :version,
                :url,
                (SELECT rowid FROM lexicons WHERE id=:id AND version=:version))
    """
    params = []
    for dep in lexicon.get("requires", []):
        param_dict = dict(dep)
        param_dict.setdefault("url", None)
        param_dict["lid"] = lexid
        params.append(param_dict)
    if params:
        cur.executemany(query.format(table="lexicon_dependencies"), params)

    if lexicon.get("extends"):
        lexicon = cast(lmf.LexiconExtension, lexicon)
        param_dict = dict(lexicon["extends"])
        param_dict.setdefault("url", None)
        param_dict["lid"] = lexid
        cur.execute(query.format(table="lexicon_extensions"), param_dict)
        extid = cur.execute(
            "SELECT rowid FROM lexicons WHERE id=? AND version=?",
            (param_dict["id"], param_dict["version"]),
        ).fetchone()[0]
    else:
        extid = lexid

    return lexid, extid


_LexIdMap = Dict[str, int]


def _build_lexid_map(lexicon: _AnyLexicon, lexid: int, extid: int) -> _LexIdMap:
    """Build a mapping of entity IDs to extended lexicon rowid."""
    lexidmap: _LexIdMap = {}
    if lexid != extid:
        lexidmap.update((e["id"], extid) for e in _entries(lexicon) if _is_external(e))
        lexidmap.update(
            (s["id"], extid)
            for e in _entries(lexicon)
            for s in _senses(e)
            if _is_external(s)
        )
        lexidmap.update(
            (ss["id"], extid) for ss in _synsets(lexicon) if _is_external(ss)
        )
    return lexidmap


T = TypeVar("T")


def _batch(sequence: Iterable[T]) -> Iterator[List[T]]:
    it = iter(sequence)
    batch = list(islice(it, 0, BATCH_SIZE))
    while len(batch):
        yield batch
        batch = list(islice(it, 0, BATCH_SIZE))


def _insert_synsets(
    synsets: Sequence[_AnySynset],
    lexid: int,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Synsets")
    # synsets
    ss_query = f"""
        INSERT INTO synsets
        VALUES (null,?,?,(SELECT rowid FROM ilis WHERE id=?),?,?,({LEXFILE_QUERY}),?)
    """
    # presupposed ILIs
    pre_ili_query = f"""
        INSERT OR IGNORE INTO ilis
        VALUES (null,?,({ILISTAT_QUERY}),?,?)
    """
    # proposed ILIs
    pro_ili_query = """
        INSERT INTO proposed_ilis
        VALUES (null,
               (SELECT ss.rowid
                  FROM synsets AS ss
                 WHERE ss.id=? AND lexicon_rowid=?),
               ?,
               ?)
    """

    for batch in _batch(_local_synsets(synsets)):

        # first add presupposed ILIs
        pre_ili_data = []
        for ss in batch:
            ili = ss["ili"]
            if ili and ili != "in":
                defn = ss.get("ili_definition")  # normally null
                text = defn["text"] if defn else None
                meta = defn["meta"] if defn else None
                pre_ili_data.append((ili, "presupposed", text, meta))
        cur.executemany(pre_ili_query, pre_ili_data)

        # then add synsets
        ss_data = (
            (
                ss["id"],
                lexid,
                ss["ili"] if ss["ili"] and ss["ili"] != "in" else None,
                ss["partOfSpeech"],
                ss.get("lexicalized", True),
                ss.get("lexfile"),
                ss["meta"],
            )
            for ss in batch
        )
        cur.executemany(ss_query, ss_data)

        # finally add proposed ILIs
        pro_ili_data = []
        for ss in batch:
            ili = ss["ili"]
            if ili == "in":
                defn = ss.get("ili_definition")
                text = defn["text"] if defn else None
                meta = defn["meta"] if defn else None
                pro_ili_data.append((ss["id"], lexid, text, meta))
        cur.executemany(pro_ili_query, pro_ili_data)

        progress.update(len(batch))


def _insert_synset_definitions(
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Definitions")
    query = f"""
        INSERT INTO definitions
        VALUES (null,?,({SYNSET_QUERY}),?,?,({SENSE_QUERY}),?)
    """
    for batch in _batch(synsets):
        data = [
            (
                lexid,
                synset["id"],
                lexidmap.get(synset["id"], lexid),
                definition["text"],
                definition.get("language"),
                definition.get("sourceSense"),
                lexidmap.get(definition.get("sourceSense", ""), lexid),
                definition["meta"],
            )
            for synset in batch
            for definition in synset.get("definitions", [])
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_synset_relations(
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Synset Relations")
    query = f"""
        INSERT INTO synset_relations
        VALUES (null,?,({SYNSET_QUERY}),({SYNSET_QUERY}),({RELTYPE_QUERY}),?)
    """
    for batch in _batch(synsets):
        data = [
            (
                lexid,
                synset["id"],
                lexidmap.get(synset["id"], lexid),
                relation["target"],
                lexidmap.get(relation["target"], lexid),
                relation["relType"],
                relation["meta"],
            )
            for synset in batch
            for relation in synset.get("relations", [])
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_entries(
    entries: Sequence[_AnyEntry],
    lexid: int,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Words")
    query = "INSERT INTO entries VALUES (null,?,?,?,?)"
    for batch in _batch(_local_entries(entries)):
        data = (
            (entry["id"], lexid, entry["lemma"]["partOfSpeech"], entry["meta"])
            for entry in batch
        )
        cur.executemany(query, data)
        progress.update(len(batch))


def _insert_forms(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Word Forms")
    query = f"INSERT INTO forms VALUES (null,?,?,({ENTRY_QUERY}),?,?,?,?)"
    for batch in _batch(entries):
        forms: List[
            Tuple[Optional[str], int, str, int, str, Optional[str], Optional[str], int]
        ] = []
        for entry in batch:
            eid = entry["id"]
            lid = lexidmap.get(eid, lexid)
            if not _is_external(entry):
                entry = cast(lmf.LexicalEntry, entry)
                written_form = entry["lemma"]["writtenForm"]
                norm = normalize_form(written_form)
                forms.append(
                    (
                        None,
                        lexid,
                        eid,
                        lid,
                        written_form,
                        norm if norm != written_form else None,
                        entry["lemma"].get("script"),
                        0,
                    )
                )
            for i, form in enumerate(_forms(entry), 1):
                if _is_external(form):
                    continue
                form = cast(lmf.Form, form)
                written_form = form["writtenForm"]
                norm = normalize_form(written_form)
                forms.append(
                    (
                        form.get("id"),
                        lexid,
                        eid,
                        lid,
                        written_form,
                        norm if norm != written_form else None,
                        form.get("script"),
                        i,
                    )
                )
        cur.executemany(query, forms)
        progress.update(len(forms))


def _insert_pronunciations(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Pronunciations")
    query = f"INSERT INTO pronunciations VALUES (({FORM_QUERY}),?,?,?,?,?)"
    for batch in _batch(entries):
        prons: List[
            Tuple[
                str,
                int,
                Optional[str],
                int,
                str,
                Optional[str],
                Optional[str],
                bool,
                Optional[str],
            ]
        ] = []
        for entry in batch:
            eid = entry["id"]
            lid = lexidmap.get(eid, lexid)
            if entry.get("lemma"):
                for p in entry["lemma"].get("pronunciations", []):
                    prons.append(
                        (
                            eid,
                            lid,
                            None,
                            0,
                            p["text"],
                            p.get("variety"),
                            p.get("notation"),
                            p.get("phonemic", True),
                            p.get("audio"),
                        )
                    )
            for i, form in enumerate(_forms(entry), 1):
                # rank is not valid in FORM_QUERY for external forms
                rank = -1 if _is_external(form) else i
                for p in form.get("pronunciations", []):
                    prons.append(
                        (
                            eid,
                            lid,
                            form.get("id"),
                            rank,
                            p["text"],
                            p.get("variety"),
                            p.get("notation"),
                            p.get("phonemic", True),
                            p.get("audio"),
                        )
                    )
        cur.executemany(query, prons)
        progress.update(len(prons))


def _insert_tags(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Word Form Tags")
    query = f"INSERT INTO tags VALUES (({FORM_QUERY}),?,?)"
    for batch in _batch(entries):
        tags: List[Tuple[str, int, Optional[str], int, str, str]] = []
        for entry in batch:
            eid = entry["id"]
            lid = lexidmap.get(eid, lexid)
            if entry.get("lemma"):
                for tag in entry["lemma"].get("tags", []):
                    tags.append((eid, lid, None, 0, tag["text"], tag["category"]))
            for i, form in enumerate(_forms(entry), 1):
                # rank is not valid in FORM_QUERY for external forms
                rank = -1 if _is_external(form) else i
                for tag in form.get("tags", []):
                    tags.append(
                        (eid, lid, form.get("id"), rank, tag["text"], tag["category"])
                    )
        cur.executemany(query, tags)
        progress.update(len(tags))


def _insert_senses(
    entries: Sequence[_AnyEntry],
    synsets: Sequence[_AnySynset],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Senses")
    ssrank = {
        s: i
        for ss in _local_synsets(synsets)
        for i, s in enumerate(ss.get("members", []))
    }
    query = f"""
        INSERT INTO senses
        VALUES (null,
                ?,
                ?,
                ({ENTRY_QUERY}),
                ?,
                ({SYNSET_QUERY}),
                ?,
                ?,
                ?)
    """
    for batch in _batch(entries):
        data = [
            (
                sense["id"],
                lexid,
                entry["id"],
                lexidmap.get(entry["id"], lexid),
                i,
                sense["synset"],
                lexidmap.get(sense["synset"], lexid),
                ssrank.get(sense["id"], DEFAULT_MEMBER_RANK),
                sense.get("lexicalized", True),
                sense["meta"],
            )
            for entry in batch
            for i, sense in enumerate(_local_senses(_senses(entry)))
        ]
        cur.executemany(query, data)
        progress.update(len(data))


def _insert_adjpositions(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
):
    progress.set(status="Sense Adjpositions")
    data = [
        (s["id"], lexidmap.get(s["id"], lexid), s["adjposition"])
        for e in entries
        for s in _local_senses(_senses(e))
        if s.get("adjposition")
    ]
    query = f"INSERT INTO adjpositions VALUES (({SENSE_QUERY}),?)"
    cur.executemany(query, data)


def _insert_counts(
    entries: Sequence[_AnyEntry],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Counts")
    data = [
        (
            lexid,
            sense["id"],
            lexidmap.get(sense["id"], lexid),
            count["value"],
            count["meta"],
        )
        for entry in entries
        for sense in _senses(entry)
        for count in sense.get("counts", [])
    ]
    query = f"INSERT INTO counts VALUES (null,?,({SENSE_QUERY}),?,?)"
    cur.executemany(query, data)
    progress.update(len(data))


def _collect_frames(lexicon: _AnyLexicon) -> List[lmf.SyntacticBehaviour]:
    # WN-LMF 1.0 syntactic behaviours are on lexical entries, and in
    # WN-LMF 1.1 they are at the lexticon level with IDs. This
    # function normalizes the two variants.

    # IDs are not required and frame strings must be unique in a
    # lexicon, so lookup syntactic behaviours by the frame string
    synbhrs: Dict[str, lmf.SyntacticBehaviour] = {
        frame["subcategorizationFrame"]: {
            "id": frame["id"],
            "subcategorizationFrame": frame["subcategorizationFrame"],
            "senses": frame.get("senses", []),
        }
        for frame in lexicon.get("frames", [])
    }
    # all relevant senses are collected into the 'senses' key
    id_senses_map = {sb["id"]: sb["senses"] for sb in synbhrs.values() if sb.get("id")}
    for entry in _entries(lexicon):
        # for WN-LMF 1.1
        for sense in _local_senses(_senses(entry)):
            for sbid in sense.get("subcat", []):
                id_senses_map[sbid].append(sense["id"])
        # for WN-LMF 1.0
        if _is_external(entry) or not entry.get("frames"):
            continue
        entry = cast(lmf.LexicalEntry, entry)
        all_senses = [s["id"] for s in _senses(entry)]
        for frame in entry.get("frames", []):
            subcat_frame = frame["subcategorizationFrame"]
            if subcat_frame not in synbhrs:
                synbhrs[subcat_frame] = {
                    "subcategorizationFrame": subcat_frame,
                    "senses": [],
                }
            senses = frame.get("senses", []) or all_senses
            synbhrs[subcat_frame]["senses"].extend(senses)
    return list(synbhrs.values())


def _insert_syntactic_behaviours(
    synbhrs: Sequence[lmf.SyntacticBehaviour],
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Syntactic Behaviours")

    query = "INSERT INTO syntactic_behaviours VALUES (null,?,?,?)"
    sbdata = [
        (sb.get("id") or None, lexid, sb["subcategorizationFrame"]) for sb in synbhrs
    ]
    cur.executemany(query, sbdata)

    # syntactic behaviours don't have a required ID; index on frame
    framemap: Dict[str, List[str]] = {
        sb["subcategorizationFrame"]: sb.get("senses", []) for sb in synbhrs
    }
    query = f"""
        INSERT INTO syntactic_behaviour_senses
        VALUES ((SELECT rowid
                   FROM syntactic_behaviours
                  WHERE lexicon_rowid=? AND frame=?),
                ({SENSE_QUERY}))
    """
    sbsdata = [
        (lexid, frame, sid, lexidmap.get(sid, lexid))
        for frame in framemap
        for sid in framemap[frame]
    ]
    cur.executemany(query, sbsdata)

    progress.update(len(synbhrs))


def _insert_sense_relations(
    lexicon: _AnyLexicon,
    lexid: int,
    lexidmap: _LexIdMap,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Sense Relations")
    # need to separate relations into those targeting senses vs synsets
    synset_ids = {ss["id"] for ss in _synsets(lexicon)}
    sense_ids = {s["id"] for e in _entries(lexicon) for s in _senses(e)}
    s_s_rels = []
    s_ss_rels = []
    for entry in _entries(lexicon):
        for sense in _senses(entry):
            slid = lexidmap.get(sense["id"], lexid)
            for relation in sense.get("relations", []):
                target_id = relation["target"]
                tlid = lexidmap.get(target_id, lexid)
                if target_id in sense_ids:
                    s_s_rels.append((sense["id"], slid, tlid, relation))
                elif target_id in synset_ids:
                    s_ss_rels.append((sense["id"], slid, tlid, relation))
                else:
                    raise wn.Error(
                        f"relation target is not a known sense or synset: {target_id}"
                    )
    hyperparams = [
        ("sense_relations", SENSE_QUERY, s_s_rels),
        ("sense_synset_relations", SYNSET_QUERY, s_ss_rels),
    ]
    for table, target_query, rels in hyperparams:
        query = f"""
            INSERT INTO {table}
            VALUES (null,?,({SENSE_QUERY}),({target_query}),({RELTYPE_QUERY}),?)
        """
        for batch in _batch(rels):
            data = [
                (
                    lexid,
                    sense_id,
                    slid,
                    relation["target"],
                    tlid,
                    relation["relType"],
                    relation["meta"],
                )
                for sense_id, slid, tlid, relation in batch
            ]
            cur.executemany(query, data)
            progress.update(len(data))


def _insert_examples(
    objs: Sequence[Union[lmf.Sense, lmf.ExternalSense, lmf.Synset, lmf.ExternalSynset]],
    lexid: int,
    lexidmap: _LexIdMap,
    table: str,
    cur: sqlite3.Cursor,
    progress: ProgressHandler,
) -> None:
    progress.set(status="Examples")
    if table == "sense_examples":
        query = f"INSERT INTO {table} VALUES (null,?,({SENSE_QUERY}),?,?,?)"
    else:
        query = f"INSERT INTO {table} VALUES (null,?,({SYNSET_QUERY}),?,?,?)"
    for batch in _batch(objs):
        data = [
            (
                lexid,
                obj["id"],
                lexidmap.get(obj["id"], lexid),
                example["text"],
                example.get("language"),
                example["meta"],
            )
            for obj in batch
            for example in obj.get("examples", [])
        ]
        # be careful of SQL injection here
        cur.executemany(query, data)
        progress.update(len(data))


def _add_ili(
    source: Path,
    progress: ProgressHandler,
) -> None:
    query = f"""
        INSERT INTO ilis
        VALUES (null,?,({ILISTAT_QUERY}),?,null)
            ON CONFLICT(id) DO
               UPDATE SET status_rowid=excluded.status_rowid,
                          definition=excluded.definition
    """
    with connect() as conn:
        cur = conn.cursor()

        progress.flash(f"Reading ILI file: {source!s}")
        ili = list(_ili.load(source))

        progress.flash("Updating ILI Status Names")
        statuses = set(info.get("status", "active") for info in ili)
        cur.executemany(
            "INSERT OR IGNORE INTO ili_statuses VALUES (null,?)",
            [(stat,) for stat in sorted(statuses)],
        )

        progress.set(count=0, total=len(ili), status="ILI")
        for batch in _batch(ili):
            data = [
                (info["ili"], info.get("status", "active"), info.get("definition"))
                for info in batch
            ]
            cur.executemany(query, data)
            progress.update(len(data))


def remove(
    lexicon: str, progress_handler: Optional[Type[ProgressHandler]] = ProgressBar
) -> None:
    """Remove lexicon(s) from the database.

    The *lexicon* argument is a :ref:`lexicon specifier
    <lexicon-specifiers>`. Note that this removes a lexicon and not a
    project, so the lexicons of projects containing multiple lexicons
    will need to be removed individually or, if applicable, a star
    specifier.

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.

    >>> wn.remove('ewn:2019')  # removes a single lexicon
    >>> wn.remove('*:1.3+omw')  # removes all lexicons with version 1.3+omw

    """
    if progress_handler is None:
        progress_handler = ProgressHandler
    progress = progress_handler(message="Removing", unit="\be5 operations")

    conn = connect()
    conn.set_progress_handler(progress.update, 100000)
    try:
        for rowid, id, _, _, _, _, version, *_ in find_lexicons(lexicon=lexicon):
            extensions = _find_all_extensions(rowid)

            with conn:

                for ext_id, ext_spec in reversed(extensions):
                    progress.set(status=f"{ext_spec} (extension)")
                    conn.execute("DELETE from lexicons WHERE rowid = ?", (ext_id,))
                    progress.flash(f"Removed {ext_spec}\n")

                spec = f"{id}:{version}"
                extra = f" (and {len(extensions)} extension(s))" if extensions else ""
                progress.set(status=f"{spec}", count=0)
                conn.execute("DELETE from lexicons WHERE rowid = ?", (rowid,))
                progress.flash(f"Removed {spec}{extra}\n")

    finally:
        progress.close()
        conn.set_progress_handler(None, 0)


def _find_all_extensions(rowid: int) -> List[Tuple[int, str]]:
    exts: List[Tuple[int, str]] = []
    for ext_id in get_lexicon_extensions(rowid):
        lexinfo = get_lexicon(ext_id)
        exts.append((ext_id, f"{lexinfo[1]}:{lexinfo[6]}"))
    return exts


def _entries(lex: _AnyLexicon) -> Sequence[_AnyEntry]:
    return lex.get("entries", [])


def _forms(e: _AnyEntry) -> Sequence[_AnyForm]:
    return e.get("forms", [])


def _senses(e: _AnyEntry) -> Sequence[_AnySense]:
    return e.get("senses", [])


def _synsets(lex: _AnyLexicon) -> Sequence[_AnySynset]:
    return lex.get("synsets", [])


def _is_external(
    x: Union[_AnyForm, _AnyLemma, _AnyEntry, _AnySense, _AnySynset]
) -> bool:
    return x.get("external", False) is True


def _local_synsets(synsets: Sequence[_AnySynset]) -> Iterator[lmf.Synset]:
    for ss in synsets:
        if _is_external(ss):
            continue
        yield cast(lmf.Synset, ss)


def _local_entries(entries: Sequence[_AnyEntry]) -> Iterator[lmf.LexicalEntry]:
    for e in entries:
        if _is_external(e):
            continue
        yield cast(lmf.LexicalEntry, e)


def _local_senses(senses: Sequence[_AnySense]) -> Iterator[lmf.Sense]:
    for s in senses:
        if _is_external(s):
            continue
        yield cast(lmf.Sense, s)
